// display

// the following lines is supposed(optional) to replace a single line in sketch, the #include"Esp8266MyWiFiUnified.h".
// For the following macros copied to main for easy modify. So, copy these to there and uncomment it could simply done.
#ifndef _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
#define _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
#include <ESP8266WiFi.h>

////////////////// paste the macros here (which is between IoT BEGIN/END .h) //////////////////
// ************************** IoT BEGIN .h ******************************
// ========================== Options and Defs/modify if need ==============================
// -------------------------- for AP or STA ------------------------------
// 1 for ap, 0 for sta.
#define AP_MODE 0
#define STA_MODE (!AP_MODE)

// -------------------------- for ap/sta credential ------------------------------ comment it and put it in .cpp
#if 1
const char *const WIFI_CRED[][2]={
    {"ESP_TESTAP", "87654321"}, // the first one will be used which is in AP mode.
    {"Office", "123456"},
    {"Xperia X Performance_bc10", "e848acbcfd98"},
};
#endif // if 1

// -------------------------- for part of hostname ------------------------------
#define LSERVICE_1 "display" // name
#define LPSERVICE_1 80 // port
#define LPOSERVICE_1 "tcp" // protocol

// -------------------------- for default connection names ------------------------------
// i.e., http://hostname.local/
#define HOSTNAME_BASE "esp8266-iot" // for local used.
#define HOSTNAME ((String(HOSTNAME_BASE)+"-"+LSERVICE_1).c_str()) // for local used.
#define REMOTE_HOSTNAME "esp8266-iot-internet" // for remote use.

#define UPDFILE "a.bin" // the remote webpage for updating to local is http://REMOTE_HOSTNAME.local/upd

// -------------------------- other settings ------------------------------
#define HTTPC_GET_DELAY delay(25) // do the delay after an HTTP GET.
#define HTTPC_GET_PROC_DELAY delay(5) // do some delay during an HTTP GET course.
#define HTTPC_CONN_TIMEOUT 3000 // the timeout during a GET, 3000ms.

// all the TMO values is restricted to no more than 35 minutes; see the Timeout() for detail. however,
// cascading could overcome.
#define TMO_IOT_LOOP_PERIOD 50 // IoT polling at each 50ms
#define TMO_AP_C_CONN_PERIOD 90000 // each 90 seconds to show connections by AP
#define TMO_STA_CONN_HEALTH_PERIOD 20000 // each 20 seconds to check the connection health by STA(while ping is alive)
#define TMO_WEB_UPDATE_PERIOD (1000*60*30) // each 30 minutes to find internet update; if cascade twice(no, use count) will become 15hr
#define TMO_MDNS_Q_PERIOD 120000 // each 120 seconds to query services
#define TMO_PING_PERIOD 5000 // each 5 seconds to ping AP when ping failed

#define STA_RECOVER_ALIVE 0 // set 0 if STA connection broken and do not reconnect. 1 to reconnect.

// [mDns] use either of one
#define USE_MDNS_STANDALONE // no auto hostname
//#define USE_MDNS_FULL // auto hostname and query

// [STAtion] use either of one
//#define USE_STA_NATIVE // single AP
#define USE_STA_WIFI_MULTI // multiple APs(o)

// [HTTPClient] use either of one
//#define USE_HTTPC_HTTP_CLIENT
//#define USE_HTTPC_WIFI_CLIENT
#define USE_HTTPC_WIFIHTTP_HYBRID // (o2)
//#define USE_HTTPC_WIFIHTTP_HYBRID_NONBLOCKING // (o1)

// [Multicast]
#define MULTICAST_ADDRESS "239.255.255.250"
#define MULTICAST_PORT "8888"
#define MULTICAST_BUFF_SIZE 15 // default is 16 in total. only (limited)for receiver.
#define MULTICAST_TR 1 // transmitting mode
#define MULTICAST_RV 2 // receiving mode

// -------------------------- for mdns ------------------------------
typedef struct smDnsRecord{
    const char *service;
    const unsigned port;
    const char *proto;
} smDnsRecord;

// -------------------------- for mdns services ------------------------------ comment it and put it in .cpp
#if 1
#if defined(USE_MDNS_FULL)
const smDnsRecord mdns_ss[]={ // the services this device provides; note, at least the LSERVICE_1 has already provided.
    //{"display", 80, "tcp"},
    //{"iot", 80, "tcp"},
    //{"https", 443, "tcp"},
    //{"test", 81, "tcp"},
    //{"motor", 80, "tcp"},
#if STA_MODE
    {"multicast", 8888, "udp"},
#endif // STA_MODE
};
#endif // USE_MDNS_FULL

const smDnsRecord mdns_sskn[]={ // the known services this device needs from other devices.
    //{"display", 80, "tcp"},
    //{"iot", 80, "tcp"},
    //{"https", 443, "tcp"},
    //{"test", 81, "tcp"},
    {"motor", 80, "tcp"},
    {"mngcenter", 80, "tcp"},
#if STA_MODE
    {"multicast", 8888, "udp"},
#endif // STA_MODE
};
#endif // if 1


// ************************** IoT END .h ******************************
////////////////// paste the macros here (which is between IoT BEGIN/END .h) //////////////////

#include"Esp8266MyWiFiUnified.h"
#endif // _ESP8266_MY_WIFI_UNIFIED_DISABLEH_


#include <U8g2lib.h>
#include"Esp8266MultiPwms.h"
#include"Esp8266Vars2File.h"

// note: http command: stop timer, before web update.
// currently uses white version board.
// known issues: 1) IRAM_ATTR ran out. 2) ping causes wdt-reset. 3) only for core version 2.7.4. for 3.0.0, too slower by test.
//              4) joystick sensing code is too repetitive. 5) not obj orientative. 6) local provided commands.
//              7) tc font. 8) use commands on multicast. 9) internet connection always failed.

#define VERSION_202106010008 "[[202106010008][AP/STA][ESP_TESTAP:87654321][esp8266-iot-display:80][esp8266-iot-internet][upd:a.bin]"\
            "[MDNS:FIXED][non WeMos TTGO][115200:160:26:4M:3MFS:1MSK:80M]]" // so the version number is the 3rd byte, length 12.
#define theVERSION "[20210601.0.8]"

#define VERSION_202106020009 "[[202106020009][AP/STA][ESP_TESTAP:87654321][esp8266-iot-display:80][esp8266-iot-internet][upd:a.bin]"\
            "[MDNS:FIXED][non WeMos TTGO][115200:160:26:4M:3MFS:1MSK:80M]]" // so the version number is the 3rd byte, length 12.
#define theVERSION "[20210602.0.9]"

// note: the 0.9a is specifically used.
#define VERSION_202106070009a "[[202106070009a][AP/STA][ESP_TESTAP:87654321][esp8266-iot-display:80][esp8266-iot-internet][upd:a.bin]"\
            "[MDNS:FIXED][non WeMos TTGO][115200:160:26:4M:3MFS:1MSK:80M]]" // so the version number is the 3rd byte, length 12.
#define theVERSION "[20210607.0.9a]"

// add temp history input, add assign http req target for processing remote cmds.
#define VERSION_202106110100 "[[202106110100][AP/STA][ESP_TESTAP:87654321][esp8266-iot-display:80][esp8266-iot-internet][upd:a.bin]"\
            "[MDNS:FIXED][non WeMos TTGO][115200:160:26:4M:3MFS:1MSK:80M]]" // so the version number is the 3rd byte, length 12.
#define theVERSION "[20210611.1.0]"

// based on v.1.0. add multicast_manual control.
#define VERSION_202106110100a "[[202106110100a][AP/STA][ESP_TESTAP:87654321][esp8266-iot-display:80][esp8266-iot-internet][upd:a.bin]"\
            "[MDNS:FIXED][non WeMos TTGO][115200:160:26:4M:3MFS:1MSK:80M]]" // so the version number is the 3rd byte, length 12.
#define theVERSION "[20210611.1.0a]"

// add chess board commands; #16.
#define VERSION_202106130101 "[[202106130101][AP/STA][ESP_TESTAP:87654321][esp8266-iot-display:80][esp8266-iot-internet][upd:a.bin]"\
            "[MDNS:FIXED][non WeMos TTGO][115200:160:26:4M:3MFS:1MSK:80M]]" // so the version number is the 3rd byte, length 12.
#define theVERSION "[20210613.1.1]"

// special version 1.1a.
#define VERSION_202106130101a "[[202106130101a][AP/STA][ESP_TESTAP:87654321][esp8266-iot-display:80][esp8266-iot-internet][upd:a.bin]"\
            "[MDNS:FIXED][non WeMos TTGO][115200:160:26:4M:3MFS:1MSK:80M]]" // so the version number is the 3rd byte, length 12.
#define theVERSION "[20210613.1.1a]"

// special version 1.1b.
#define VERSION_202107190101b "[[202107190101b][AP/STA][ESP_TESTAP:87654321][esp8266-iot-display:80][esp8266-iot-internet][upd:a.bin]"\
            "[MDNS:FIXED][non WeMos TTGO][115200:160:26:4M:3MFS:1MSK:80M]]" // so the version number is the 3rd byte, length 12.
#define theVERSION "[20210719.1.1b]"
// revised from v.1.1a.
// modified TMO_IOT_LOOP_PERIOD for better responses of remote http requests.
// exclusively for gesture; add gesture response http-commands.


unsigned stop_the_timer=0;
unsigned poweroff=0;
unsigned pwron_keep_alive=0;
unsigned copy_sta_do_the_connect; // default is 0. for STA connection. since order in main() is of concern, so copy.

// [[main controls]]
String web_link, speed_str;
String netstr_ssid, netstr_localip, netstr_remoteip, netstr_mngcenter_ip, netstr_mngcenter_web;
unsigned pwron_spd_conn=0;
String multicast_tr_count="-", multicast_found="";
unsigned motor_issued_cmd_state=0;
unsigned motor_http_request_passthrough=0;
String motor_issued_cmd_str="";
char motor_devs_ips[4][17];
String copy_http_got_str="manual http request\nreturned strings\nentire display\n";
bool only_pause_the_auto_http_req=true; // default true means not pause.
int copy_multicast_is_broadcasting=-1; // do have the re-request mechanism, options to be (3, off, rereq), (4, tr, rereq), (5, rv, rereq).
unsigned multicast_manual=0; // when manual control is enabled while multicast transmitting.
unsigned slowdown=1000; // 1000ms to slow down the function my_http_request() polling.


// ------------------------------------ remote http commands
const char *http_commands_str[]={ // between "< >" for replacement, ";" for delimit.
    "<ANY URL>; e.g., http://tw.yahoo.com/",
    "<SET TARGET>; e.g., 192.168.4.10, must ip only",
    "<CUSTOM PAGE>; e.g., /a_custom_page",

    "/webupdinfo?ip=<remote ip address>;",
    "/webupdinfo?addr=<remote full url>; e.g., http://192.168.1.1/x.bin",
    "/newwebupdinfo?name=<remote website>&port=<port>;",
    "/sethttprequestip?ip=<remote ip address>;",
    "/download?filename=<name>;",
    "/setupdfile?filename=<name>;",
    "/del?filename=<name>;",
    "/cp?src=<src_filename>&dst=<dst_filename>;",
    "/params/mtr/loadparams?id=<id>;",
    "/params/mtr/deleteparams?id=<id>;",
    "/t?t=<0(disable)1(open)2(close)3(inc opening)4(dec closing)5(hold)>;",
    "/params/mtr/sync?select=<0(native) 1(mild) 2(gentle) 3(wild) 4(20sW) 5(60sW)>;",
    "/multicast?cmd=<0(disable) 2(as receiver). for sta mode only, multicast/8888/udp>;",
    "/multicastassign?name=<mngcenter or display. still in RV mode>;",

    "/listcmds",
    "/update",
    "/login",
    "/upd",
    "/ip",
    "/forcegblupdate",
    "/upload",
    "/mdns",
    "/version",
    "/ls",
    "/reset",
    "/factory/loaddefault",
    "/factory/casttodefault",
    "/factory/format",
    "/factory/emptydb",
    "/factory/emptyfac",
    "/factory/renew",
    "/factory/debug/dump",
    "/params/mtr/listparams",
    "/params/mtr/unlockparams",
    "/params/mtr/probe",
    "/params/mtr/writeback",
    "/params/mtr/on",
    "/params/mtr/off",
    "/params/mtr/syncon",
    "/params/mtr/syncoff",
    "/params/mtr/speedbypass",
    "/params/mtr/test",
    "/params/gate/open",
    "/params/gate/looseopen",
    "/params/gate/close",
    "/params/gate/looseclose",
    "/params/gate/efficiency",
    "/params/gate/performance",
    "/factory/mtr/reverse",
    "/multicastinfo",

    "CHOOSE THE TARGET",
    "CANCEL", "-", "-", "-", // need extra 3 options at head, and a device pick, a cancel and more 3 lines in order to not excess boundary.
};

const unsigned char http_commands_need_params[17]={ // in http_commands_str[] where needs the number of params.
    1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1
};

/*IRAM_ATTR */const char *const remote_command_strs[4][4]={
    "CANCEL", "sync 0", "sync 1", "sync 2",
    "sync 3", "sync 4", "sync 5", "mtr on",
    " open", " close", "i-open", "d-cloz",
    "mtroff", "hreq.2", "hreq.3", "hreq.4",
};

const char *remote_fast_commands[4][4]={
    "",
    "/params/mtr/sync?select=0",
    "/params/mtr/sync?select=1",
    "/params/mtr/sync?select=2",
    "/params/mtr/sync?select=3",
    "/params/mtr/sync?select=4",
    "/params/mtr/sync?select=5",
    "/params/mtr/on",
    "/t?t=1",
    "/t?t=2",
    "/t?t=3",
    "/t?t=4",
    "/params/mtr/off",
    "/sethttprequestip?ip=192.168.4.2",
    "/sethttprequestip?ip=192.168.4.3",
    "/sethttprequestip?ip=192.168.4.4",
};


// ------------------------------------ default history strings
const char *input_str_history_default[]={
    "http://",
    "esp8266-iot",
    ".local",
    "esp8266-iot.local",
    "esp8266-iot-internet",
    "esp8266-iot-s2-1-motor.local",
    "esp8266-iot-s2-2-motor.local",
    "esp8266-iot-mngcenter.local",
    "esp8266-",
    "tcp",
    "udp",
    "ip",
    "mdns",
    "www.",
    "com.",
    "tw/",
    ".com.tw/",
    "192.168.",
    "192.168.4.1",
    "192.168.4.2",
    "192.168.4.3",
    "192.168.4.4",
    "192.168.4.5",
    "192.168.4.6",
    "motor",
    "display",
    "mngcenter",
    "multicast",
    "multicastinfo",
    "80",
    "8888",
    "239.255.255.250",
    "a.bin",
    "b.bin",
    "c.bin",
};


// ------------------------------------ timer settings
#define DRAW_INT_GPIO           12
#define DRAW_TRG_GPIO           13
#define DRAW_PERIOD             80000 // 80ms. less than 80ms will wdt-reset when sta mode.
#define BTN_PERIOD              20000 // 20ms. less than 20ms will wdt-reset when sta mode.


// ------------------------ mux
// mux s2=gnd, s1=gpio13, s0=gpio12, use a0, a1, a2, a3 for y, x, button, battery.

//#define MUX_S0_GPIO             16 // TTGO
#define MUX_S0_GPIO             2 // non TTGO
#define MUX_S1_GPIO             3

//#define BUTTON_GPIO             5 // TTGO
#define BUTTON_GPIO             14 // non TTGO

#define VOLTAGE_REAL            14.5f
#define VOLTAGE_TRAN            3.248f
#define VOFFSET                 17
#define BVOLTAGE(x)             (((x)>VOFFSET? (x)-17: 0)*(3.3f/VOLTAGE_TRAN*VOLTAGE_REAL/1024.f))


// ------------------------------------ u8g2 config
// TTGO(black board)
//#define OLED_SDA  2
//#define OLED_SCL 14
//#define OLED_RST  4

// non TTGO(white board)
#define OLED_SDA  4
#define OLED_SCL  5
#define OLED_RST 16

U8G2_SSD1306_128X32_UNIVISION_F_SW_I2C u8g2(U8G2_R0, OLED_SCL, OLED_SDA, OLED_RST);

void setup_u8g2(){
    u8g2.begin();
}


// ------------------------------------ timer processing for both joystick and u8g2 drawing.

// to store draw functions and for traversal.
// each linkedlist draws a page(of course the present page), each node draws parts of the page. order is of concerned.
typedef struct sFooLinkedList{
    void (*fp)();
    sFooLinkedList *next;
    sFooLinkedList(): fp(0), next(0){;};
    ~sFooLinkedList(){delete this->next;};
} sFooLinkedList;

cHwTimer timer_draw_manip;                  // the timer obj. be aware that hwtimer1 must bring up first.
IRAM_ATTR unsigned timer_trigger=0;         // the variable for toggling.
IRAM_ATTR void isr_draw_manip_timer(){digitalWrite(DRAW_TRG_GPIO, timer_trigger^=1);} // the timer isr.

IRAM_ATTR sFooLinkedList *draw_page_1=0;    // the page presently.
IRAM_ATTR sFooLinkedList *draw_page_2=0;    // the incoming page. in order for page sliding animation.
/**IRAM_ATTR **/int dpox=0; // the draw page offset, x. in order for page sliding animation. use IRAM_ATTR causes u8g2.drawBox fault.
/**IRAM_ATTR **/int dpoy=0; // the draw page offset, y. in order for page sliding animation. use IRAM_ATTR causes u8g2.drawBox fault.
IRAM_ATTR int slide_points_x[]={5, 16, 28, 48, 20, 6, 5};   // for slide x animation.
IRAM_ATTR int slide_points_y[]={2, 3, 6, 15, 3, 2, 1};      // for slide y animation.
IRAM_ATTR int slide_idx;    // keeps the index while animating.
#define SLIDE_DIR_TIPIDLE   0
#define SLIDE_DIR_TIPLEFT   1
#define SLIDE_DIR_TIPRIGHT  -1
#define SLIDE_DIR_TIPDOWN   -2
#define SLIDE_DIR_TIPUP     2
IRAM_ATTR int slide_dir;    // keeps the direction while animating. (-2, down), (-1, right), (1, left), (2, up).
IRAM_ATTR int slide_2_pos;  // it mainly presenting the zero position of the incoming page. nonzero means still animating.
#define DRAW_HOLDON_NEGATIVE    0 // idle, no hold on.
#define DRAW_HOLDON_ISSUED      1 // starting hold on.
#define DRAW_HOLDON_ING         2 // processing.
#define DRAW_HOLDON_RESPED      3 // responsed.
#define DRAW_HOLDON_RESP_TMO    3500 // timeout value 3500ms for waiting response.
#define DRAW_HOLDON_DRAW_TMO    1500 // time 1500ms for drawing response.
void (*DrawHoldOn_resp_fptr)()=0; // responsed function pointer of a draw.
unsigned DrawHoldOn_state=0;  // a stater for hold on drawing.
IRAM_ATTR unsigned DrawSubmit=0;            // set 1 to start drawing/refresh.
IRAM_ATTR unsigned AutoDrawSubmit=0;        // set 1 to start auto drawing/refresh.

IRAM_ATTR unsigned joystick_vol_old;        // the previously battery voltage adc value.
IRAM_ATTR unsigned joystick[4];             // 4 channels ADC values. (0, y), (1, x), (2, button), (3, battery).
IRAM_ATTR unsigned joystick_tri[4];         // 4 channels ADC values of changes. if changes that channel has triggered.
IRAM_ATTR unsigned joystick_alive=0;        // any channel has triggered is alive. all channels unchanged is not alive.
unsigned joystick_x_old, joystick_y_old, joystick_b_old; // x, y, batt. used for probing being triggered(changed) or not.
#define JS_DIR_IDLE         0
#define JS_DIR_LEFT         1
#define JS_DIR_RIGHT        2
#define JS_DIR_DOWN         4
#define JS_DIR_UP           8
#define JS_DIR_PUSH         16
unsigned joystick_dir; // idle=0, left=1, right=2, down=4, up=8, push=16.
unsigned joystick_action; // (0, idle); (1, under acting).

void DrawSetup(){ // for timer isr and gpio isr. initialize once.
    pinMode(DRAW_TRG_GPIO, OUTPUT); // this gpio is to trigger intr gpio.
    digitalWrite(DRAW_TRG_GPIO, LOW);
    attachInterrupt(digitalPinToInterrupt(DRAW_INT_GPIO), isr_draw_manip, CHANGE); // setup gpio interrupt.
}

void DrawConfig(class sFooLinkedList *&draw_link, void (*fptr)()){ // insert function pointers to which draws the page buffer.
    if (!draw_link){
        draw_link=new sFooLinkedList;
        draw_link->fp=fptr;
        return;
    }
    sFooLinkedList *a_link=draw_link;
    while (a_link->next) a_link=a_link->next;
    a_link->next=new sFooLinkedList;
    a_link->next->fp=fptr;
}

void DrawDeConfig(class sFooLinkedList *&draw_link){ // remove the tail function pointer.
    if (!draw_link) return;
    sFooLinkedList **a_link=&draw_link, *b_link=*a_link;
    while (b_link->next){
        a_link=&(b_link->next);
        b_link=b_link->next;
    }
    *a_link=0;
    delete b_link;
}

void DrawEnd(class sFooLinkedList *&draw_link){ // must called if do not draw anymore.
    // note that violations(in isr) do exist against the 3 draw functions. advanced treatment if needed. e.g., do it in isr or use gate.
    ////unsigned i=DrawSubmit, j=AutoDrawSubmit; //// this way only usually works. now rather to put it primitive.
    ////AutoDrawSubmit=0;
    ////DrawSubmit=2;
    delete draw_link;
    draw_link=0;
    ////DrawSubmit=i;
    ////AutoDrawSubmit=j;
}

IRAM_ATTR void DrawPageUpdateOffset_1(){ // handling while animating for present page.
    int i;
    dpox=dpoy=0;
    switch (slide_dir){
        case SLIDE_DIR_TIPDOWN:
            for (i=0; i<=slide_idx; i++) dpoy+=slide_points_y[i];
            dpoy=0-dpoy;
            break;
        case SLIDE_DIR_TIPRIGHT:
            for (i=0; i<=slide_idx; i++) dpox+=slide_points_x[i];
            dpox=0-dpox;
            break;
        case SLIDE_DIR_TIPIDLE: break; // null
        case SLIDE_DIR_TIPLEFT:
            for (i=0; i<=slide_idx; i++) dpox+=slide_points_x[i];
            break;
        case SLIDE_DIR_TIPUP:
            for (i=0; i<=slide_idx; i++) dpoy+=slide_points_y[i];
            break;
    }
}

IRAM_ATTR void DrawPageUpdateOffset_2(){ // handling while animating for incoming page.
    switch (slide_dir){
        case SLIDE_DIR_TIPDOWN:
            slide_2_pos-=slide_points_y[slide_idx];
            dpoy=slide_2_pos;
            break;
        case SLIDE_DIR_TIPRIGHT:
            slide_2_pos-=slide_points_x[slide_idx];
            dpox=slide_2_pos;
            break;
        case SLIDE_DIR_TIPIDLE: break; // null
        case SLIDE_DIR_TIPLEFT:
            slide_2_pos+=slide_points_x[slide_idx];
            dpox=slide_2_pos;
            break;
        case SLIDE_DIR_TIPUP:
            slide_2_pos+=slide_points_y[slide_idx];
            dpoy=slide_2_pos;
            break;
    }
    slide_idx++;
}

IRAM_ATTR void demandcall_draw_manip(){
    if (DrawSubmit==1){
        DrawSubmit=0;
        u8g2.clearBuffer();
        sFooLinkedList *a_link=draw_page_1;
        if (slide_2_pos) DrawPageUpdateOffset_1();
        while (a_link){
            if (a_link->fp) (a_link->fp)();
            a_link=a_link->next;
        }
        if (slide_2_pos){
            a_link=draw_page_2;
            DrawPageUpdateOffset_2();
            while (a_link){
                if (a_link->fp) (a_link->fp)();
                a_link=a_link->next;
            }
            if (!slide_2_pos) DrawSubmit=2; // in order to hold the refresh of animating.
        }
        u8g2.sendBuffer();
    }
    if (AutoDrawSubmit && DrawSubmit!=2) DrawSubmit=1;
}

IRAM_ATTR void isr_draw_manip(){

    // ------ for drawing
    IRAM_ATTR static unsigned t=0;
    t+=BTN_PERIOD;
    if (t>=DRAW_PERIOD){
        t=0;
        demandcall_draw_manip();
    }

    // ------ for button sensing
    // joystick sampling. CD74HC4051E Mux.
    // a0, a1, a2, a3 for y, x, button, battery.

    IRAM_ATTR static unsigned pn=0;
    digitalWrite(MUX_S0_GPIO, pn&0x1);
    digitalWrite(MUX_S1_GPIO, (pn>>1)&0x1);
    delayMicroseconds(1);
    joystick[pn]=analogRead(A0);

    if (joystick[0]){ // the joystick is now working
        switch (pn){
            case 0: // y
                if (joystick[0]<300){
                    joystick_tri[0]+=((joystick_tri[0]&0x1)? 2: 1); // odd if dir=-y.
                    joystick_alive=1;
                }
                else if (joystick[0]>700){
                    joystick_tri[0]+=((joystick_tri[0]&0x1)? 1: 2); // even if dir=+y.
                    joystick_alive=1;
                }
                break;
            case 1: // x
                if (joystick[1]<300){
                    joystick_tri[1]+=((joystick_tri[1]&0x1)? 2: 1); // odd if dir=-x.
                    joystick_alive=1;
                }
                else if (joystick[1]>700){
                    joystick_tri[1]+=((joystick_tri[1]&0x1)? 1: 2); // even if dir=+x.
                    joystick_alive=1;
                }
                break;
            case 2: // button
                if (joystick[2]>300){ // caught the make code
                    joystick_tri[2]++;
                    joystick_alive=1;
                }
                break;
            case 3: // battery voltage, low detection.
                // around the battery voltage detection of 3.45V, the LDO will output 2.63V and can not boot up.
                // so the threshold sould be 3.7V(LDO 2.9V).
                extern const char *bat_status;
                joystick[3]=joystick_vol_old*0.8f+joystick[3]*0.2f;
                joystick_vol_old=joystick[3];
                if (BVOLTAGE(joystick[3])<3.7f){ // battery low
                    if (joystick_tri[3]<1000) joystick_tri[3]++;
                    else if (joystick_tri[3]==1000){
                        bat_status="l"; // 4
                        joystick_tri[3]++;
                    }
                }
                else if (joystick_tri[3]){
                    if (joystick_tri[3]==1) bat_status="-"; // 0
                    joystick_tri[3]--;
                }
                break;
        }
    }
    if (++pn>3) pn=0;
}


// ------------------------------------ u8g2 drawing
#define U8G2_X          128
#define U8G2_Y          32
IRAM_ATTR unsigned countdown=0;
int mapx_idx, mapy_idx;
unsigned locked_lr, locked_ud;
IRAM_ATTR int jtk_pos_x, jtk_pos_y;
IRAM_ATTR unsigned input_str_insmode=0; // input insert mode. positioning from 1 to str length.
IRAM_ATTR unsigned input_str_insmode_flag=0; // input insert mode flag. flag 1 is under insert typing.
IRAM_ATTR int viewoffset_input_str=0;
/*IRAM_ATTR */String input_str, view_input_str, input_param[2]; // not for IRAM_ATTR
/*IRAM_ATTR */const char *opt_i_box[4]; // to store the strings of options.
String input_str_history, input_str_history_notsave;
IRAM_ATTR unsigned input_str_history_flag;
IRAM_ATTR unsigned input_str_history_posi;
IRAM_ATTR unsigned input_str_history_posj;
char loop_u8g2_message_str[17]; // to store the customized message of messagebox.
const char alphabet[][26]={
    "abcdefghijklm ~!@#$%^&_()",
    "nopqrstuvwxyz`+-*/=\\'?\"{}",
    "LUUCBS1234567890<>,;.:|[]"
}; // LUUCBS, line-feed, upper-once, upper-fixed, clear, backspace, save.

// ==== u8g2.setFont(u8g2_font_5x7_tr); === occupy 5x7 pixels === need not preserve empty pixel ======= 128x32 ======
// in a row, positioning start from 1, end to 121, 25 characters
// in a col, positioning start from 2(8), end to 23(29), 4 characters
// first col, (1, 8), (1, 15), (1, 22), (1, 29)

// ==== u8g2.setFont(u8g2_font_8x13B_mf); === occupy 8x12 pixels === need not preserve empty pixel ======= 128x32 ======
// in a row, positioning start from 0, end to 120, 16 characters
// in a col, positioning start from 1(13), end to 14(26), 2 characters
// first col, (0, 13), (0, 26)

// ==== u8g2.setFont(u8g2_font_unifont_t_chinese1); === 1 to 3 == 128x32 ======
// u8g2.enableUTF8Print(); first
// u8g2.setCursor(x, y);
// u8g2.print("");
// or u8g2.drawUTF8(x, y, "");

// ==== u8g2.setFont(u8g2_font_unifont_t_symbols); ===== 128x32 ======
// u8g2.enableUTF8Print(); first
// u8g2.drawUTF8(0, 28, "Snowman: ☃");

IRAM_ATTR void loop_u8g2_alphabet(){
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawStr(dpox+1,  dpoy+8, alphabet[0]);
    u8g2.drawStr(dpox+1, dpoy+15, alphabet[1]);
    u8g2.drawStr(dpox+1, dpoy+22, alphabet[2]);
    u8g2.drawBox(dpox+11, dpoy+16, 5, 7); // the LU"U"BS.
    u8g2.drawBox(dpox+103, dpoy+25, 25, 7); // the "ihHsk".
    u8g2.setColorIndex(0);
    u8g2.drawStr(dpox+11, dpoy+22, "U");
    u8g2.drawStr(dpox+103, dpoy+31, "ihHsk"); // insert mode, history-replace, history-append, save input as history, keep.
    u8g2.setColorIndex(1);
    /// could instead use this
    ///u8g2.setDrawColor(0);
    ///u8g2.drawStr(11, 22, "U");
    ///u8g2.setDrawColor(1);
    //u8g2.drawStr(103, 31, "|||||"); // u8g2.drawStr(103, 31, "nxt->"); // u8g2.drawStr(103, 31, "__|__");
}

/*IRAM_ATTR */int loop_u8g2_alphabet_change_str_window(String &dst, String &src, int cur_pos){ // transform the input_str view.
    // since in insert mode, the cursor is from 1 to str.length. so 0 represents to be non insert mode cur_pos which is str.length+1.
    // by the macro, if str.length>it, will be transformed; become "..str.." or "str.." or "..str" or "..str_".
    // so the max length to show will be it+4. recommend set it as 16.
    // and the return value is the offset(for add) for absolute cur_pos upon str to become new relative cur_pos upon new str.
    // note: limited to 1022 chars at later code; no concern to more.
    #define U8G2_INPUTSTR_LENLIMIT 16
    #define U8G2_INPUTSTR_BOXLEN (U8G2_INPUTSTR_LENLIMIT+4)
    IRAM_ATTR static int pivot; // right boundary
    IRAM_ATTR static unsigned len; // in order to detect length change as to change into better view.
    if (src.length()<=U8G2_INPUTSTR_LENLIMIT){ // none to do.
        dst="  "+src;
        len=src.length(); // need to update since in insert mode could be less than limit and then become over limit.
        return 0+2; // 2 is "  ".
    }
    else if (!cur_pos){ // normal mode.
        len=src.length(); // also need to update; as i know, when leave insert mode would be affective.
        dst=".."+src.substring(len-U8G2_INPUTSTR_LENLIMIT);
        return 0;
    }
    else { // insert mode. note that cur_pos circulates between 1 to src.length.
        unsigned srclen=src.length();
        if (cur_pos==srclen){ // initial condition /////
            len=srclen; // reset the len for later length-changed detection.
            pivot=len;
            dst=".."+src.substring(srclen-U8G2_INPUTSTR_LENLIMIT);
            return U8G2_INPUTSTR_LENLIMIT-pivot+2; // the 2 is "..".
        }
        else if (cur_pos<U8G2_INPUTSTR_LENLIMIT+1){ // left boundary condition
            // very critical note: the order of left boundary condition first then sub-initial cond.(the following else-if) is of concern.
            pivot=U8G2_INPUTSTR_LENLIMIT;
            dst="  "+src.substring(0, U8G2_INPUTSTR_LENLIMIT)+"..";
            return 0+2; // 2 is "  ".
        }
        else if (cur_pos>srclen-U8G2_INPUTSTR_LENLIMIT){ // right. ///// note: it incl. initial cond., but excluded here by else-if.
            if (len!=srclen && cur_pos<srclen){ // length change detected, in order to have better view when inserting.
                pivot=cur_pos; // note that the relative position of pivot and cur_pos does not change while inserting.
                dst=".."+src.substring(pivot-U8G2_INPUTSTR_LENLIMIT, pivot)+"..";
            }
            else {
                pivot=srclen;
                dst=".."+src.substring(srclen-U8G2_INPUTSTR_LENLIMIT);
            }
            return U8G2_INPUTSTR_LENLIMIT-pivot+2;
        }
        else { // middle condition.
            if ((cur_pos>pivot) || (cur_pos+U8G2_INPUTSTR_LENLIMIT-1<pivot) || (len!=srclen)){
                if (len!=srclen){ // length changed.
                    pivot=cur_pos; // become better view.
                }
                else pivot=cur_pos+U8G2_INPUTSTR_LENLIMIT-1;
                dst=".."+src.substring(pivot-U8G2_INPUTSTR_LENLIMIT, pivot)+"..";
            }
            /// return U8G2_INPUTSTR_LENLIMIT-pivot+2;
        }
    }
    return U8G2_INPUTSTR_LENLIMIT-pivot+2; ///
}

/*IRAM_ATTR */void loop_u8g2_alphabet_change(){
#define U8G2_POSX_INS 103 // the 1st char in enhance mode
    int a=jtk_pos_x*5+1;
    int b=(jtk_pos_y+1)*7+1;

    if (input_str_insmode) viewoffset_input_str=loop_u8g2_alphabet_change_str_window(view_input_str, input_str, input_str_insmode);
    else if (input_str_insmode_flag) viewoffset_input_str=loop_u8g2_alphabet_change_str_window(view_input_str, input_str, jtk_pos_x);
    else viewoffset_input_str=loop_u8g2_alphabet_change_str_window(view_input_str, input_str, 0);

    if (jtk_pos_y<3) u8g2.drawBox(dpox+a, dpoy+b-7, 5, 7);
    u8g2.drawBox(dpox+1, dpoy+22, 5*(U8G2_INPUTSTR_BOXLEN), 7);
    u8g2.setDrawColor(0);
    if (jtk_pos_y<3) u8g2.drawStr(dpox+a, dpoy+b, String(alphabet[jtk_pos_y][jtk_pos_x]).c_str());
    if (input_str_insmode || input_str_insmode_flag) u8g2.drawStr(dpox+1, dpoy+29, (view_input_str).c_str());
    else u8g2.drawStr(dpox+1, dpoy+29, (view_input_str+"_").c_str());
    u8g2.setDrawColor(1);
    u8g2.drawBox(dpox+103, dpoy+25, 25, 7); // the box of "ihHsk".
    u8g2.setColorIndex(0);

    ///// u8g2.drawStr(dpox+103, dpoy+31, "ihHsk"); // insert mode, history-replace, history-append, save input as history, keep.
    u8g2.drawStr(dpox+103+5, dpoy+31, "hHsk");
    if (input_str_insmode || input_str_insmode_flag) u8g2.drawStr(dpox+103, dpoy+31, "I");
    else u8g2.drawStr(dpox+103, dpoy+31, "i");

    if ((jtk_pos_y==3) && !input_str_insmode_flag) u8g2.drawBox(dpox+U8G2_POSX_INS+jtk_pos_x*5, dpoy+30, 5, 2); // enhance mode underline.
    u8g2.setColorIndex(1);
    if (input_str_insmode_flag) u8g2.drawBox(dpox+1+(jtk_pos_x-1+viewoffset_input_str)*5, dpoy+29, 5, 3); // floating string underline.
    else if (input_str_insmode) u8g2.drawBox(dpox+1+(input_str_insmode-1+viewoffset_input_str)*5, dpoy+29, 5, 3); // fixed string underline.
}

/*IRAM_ATTR */void loop_u8g2_on(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+35, dpoy+13, "NOW ON");
}

/*IRAM_ATTR */void loop_u8g2_off(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+35, dpoy+26, "NOW OFF");
}

/*IRAM_ATTR */void loop_u8g2_success(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawBox(33, 0, 60, U8G2_Y);
    u8g2.setDrawColor(0);
    u8g2.drawStr(35, 23, "SUCCESS");
    u8g2.setDrawColor(1);
}

/*IRAM_ATTR */void loop_u8g2_failure(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawBox(33, 0, 60, U8G2_Y);
    u8g2.setDrawColor(0);
    u8g2.drawStr(35, 23, "FAILURE");
    u8g2.setDrawColor(1);
}

/*IRAM_ATTR */void loop_u8g2_no_resp(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawBox(33, 0, 60, U8G2_Y);
    u8g2.setDrawColor(0);
    u8g2.drawStr(35, 23, "NO RESP");
    u8g2.setDrawColor(1);
}

/*IRAM_ATTR */void loop_u8g2_message(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawBox(0, 10, U8G2_X, 15);
    u8g2.setDrawColor(0);
    u8g2.drawStr(0, 23, loop_u8g2_message_str);
    u8g2.setDrawColor(1);
}

/*IRAM_ATTR */void loop_u8g2_options_3_boxes(){ // horizontal
    // box offset=(0, 8, 42, 9), (41, 8, 42, 9), (82, 8, 42, 9)
    // char offset=(1, 15), (42, 15), (83, 15)
    unsigned b[]={0, 41, 82};
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawStr(dpox+1, dpoy+15, opt_i_box[0]);
    u8g2.drawStr(dpox+42, dpoy+15, opt_i_box[1]);
    u8g2.drawStr(dpox+83, dpoy+15, opt_i_box[2]);
    u8g2.drawBox(dpox+b[jtk_pos_x], dpoy+8, 42, 9);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+b[jtk_pos_x]+1, dpoy+15, opt_i_box[jtk_pos_x]);
    u8g2.setDrawColor(1);
}

/*IRAM_ATTR */void loop_u8g2_options_4_boxes(){ // vertical
    // box offset=(0, 0, 128, 8), (0, 8, 128, 8), (0, 16, 128, 8), (0, 24, 128, 8)
    // char offset=(1, 8), (1, 15), (1, 22), (1, 29)
    unsigned b[]={0, 8, 16, 24, 32};
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawStr(dpox+1, dpoy+8, opt_i_box[0]);
    u8g2.drawStr(dpox+1, dpoy+15, opt_i_box[1]);
    u8g2.drawStr(dpox+1, dpoy+22, opt_i_box[2]);
    u8g2.drawStr(dpox+1, dpoy+29, opt_i_box[3]);
    u8g2.drawBox(dpox+0, dpoy+b[jtk_pos_y], 128, 8);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+1, dpoy+b[jtk_pos_y+1]-1, opt_i_box[jtk_pos_y]);
    u8g2.setDrawColor(1);
}

/*IRAM_ATTR */void loop_u8g2_options_16_boxes_prepare(){ // chess board
    // void U8G2::drawFrame(u8g2_uint_t x, u8g2_uint_t y, u8g2_uint_t w, u8g2_uint_t h)
    // void U8G2::drawLine(u8g2_uint_t x0, u8g2_uint_t y0, u8g2_uint_t x1, u8g2_uint_t y1)
    //
    // row: 1 element 6 chars, all 120 pixels. 5 pixels of v-frame lines. starts from (l=2, c=3), end to (c=125, l=126).
    // col: 1 element 1 char, all 28 pixels. 4 pixels of h-frame lines, top has none. starts from (c=0, l=7) end to (c=30, l=31).
    // fill box should be w=30, h=7. start from (x=3, y=0), end to (x=125, y=30), skip=1.
    // line points: x=2, 33, 64, 95, 126; y=0, 7, 15, 23, 31.

    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawLine(2, 0, 2, 31); // v1
    u8g2.drawLine(33, 0, 33, 31); // v2
    u8g2.drawLine(64, 0, 64, 31); // v3
    u8g2.drawLine(95, 0, 95, 31); // v4
    u8g2.drawLine(126, 0, 126, 31); // v5
    u8g2.drawLine(2, 7, 126, 7); // h1
    u8g2.drawLine(2, 15, 126, 15); // h2
    u8g2.drawLine(2, 23, 126, 23); // h3
    u8g2.drawLine(2, 31, 126, 31); // h4
    for (int i=0, starty=6, skipy=8; i<4; i++, starty+=skipy){ // y
        for (int j=0, startx=3, skipx=31; j<4; j++, startx+=skipx){ // x
            u8g2.drawStr(startx, starty, remote_command_strs[i][j]);
        }
    }
}

/*IRAM_ATTR */void loop_u8g2_options_16_boxes(){ // chess board
    /*IRAM_ATTR */static int x=jtk_pos_x, y=jtk_pos_y;
    u8g2.setDrawColor(0);
    u8g2.drawBox(3+31*x, 0+8*y, 30, 7);
    u8g2.setDrawColor(1);
    u8g2.drawStr(3+31*x, 6+8*y, remote_command_strs[y][x]);
    x=jtk_pos_x;
    y=jtk_pos_y;
    //// u8g2.setFont(u8g2_font_6x10_tr);
    u8g2.setDrawColor(1);
    u8g2.drawBox(3+31*x, 0+8*y, 30, 7);
    u8g2.setDrawColor(0);
    u8g2.drawStr(3+31*x, 6+8*y, remote_command_strs[y][x]);
    u8g2.setDrawColor(1);
}

/*IRAM_ATTR */void loop_u8g2_welcome(){
    u8g2.setFont(u8g2_font_luBIS18_tr);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+4, dpoy+25, "welcome");
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
}

/*IRAM_ATTR */void loop_u8g2_poweroff(){
    u8g2.setFont(u8g2_font_luBIS18_tr);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+4, dpoy+25, "off now");
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
}

/*IRAM_ATTR */void loop_u8g2_settings(){ // settings
    u8g2.setFont(u8g2_font_luBIS18_tr);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+4, dpoy+25, "SETTING");
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
}

/*IRAM_ATTR */void loop_u8g2_settings_poweronconnect(){ // settings->poweronconnect
    u8g2.setFont(u8g2_font_luBIS18_tr);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+4, dpoy+25, "PrOnCnn");
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
}

/*IRAM_ATTR */void loop_u8g2_settings_poweronspeed(){ // [[main controls]] settings->poweronspeed
    u8g2.setFont(u8g2_font_luBIS18_tr);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+4, dpoy+25, "PrOnSpd");
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
}

/*IRAM_ATTR */void loop_u8g2_settings_keepalive(){ // settings->keepalive
    u8g2.setFont(u8g2_font_luBIS18_tr);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+4, dpoy+25, "KpAlive");
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
}

/*IRAM_ATTR */void loop_u8g2_settings_TOff1M(){ // settings->off_the_timer_for_1_minute
    u8g2.setFont(u8g2_font_luBIS18_tr);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+4, dpoy+25, "TOff1M");
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
}

/*IRAM_ATTR */void loop_u8g2_bgwhite(){
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
}

/*IRAM_ATTR */void loop_u8g2_bgblack(){
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, U8G2_Y);
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
}

IRAM_ATTR void loop_u8g2_status_051(){ // connection status; c/d connected/disconnected.
    u8g2.setFont(u8g2_font_5x7_tr);
#if STA_MODE
    if (!is_station_mode_established) u8g2.drawStr(103, 31, "d");
    else u8g2.drawStr(103, 31, "c");
#else // AP_MODE
    u8g2.drawStr(103, 31, "a"); // means AP mode
#endif // STA_MODE
}

IRAM_ATTR void loop_u8g2_status_052(){ // battery status. charge/discharge/full/low.
    extern const char *bat_status;
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawStr(108, 31, bat_status);
}

IRAM_ATTR void loop_u8g2_status_053(){ // in (h)ttp mode or in multicast (r)eceiver or (t)ransmitter mode.
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawStr(113, 31, multicast_is_broadcasting? ((multicast_is_broadcasting==MULTICAST_TR)? "T": "R"): "H");
}

IRAM_ATTR void loop_u8g2_status_054(){ // the number of multicast transmitters/receivers found in (multicast, udp, 8888) channel.
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawStr(118, 31, multicast_tr_count.c_str());
}

IRAM_ATTR void loop_u8g2_status_055(){ // countdown to poweroff
    u8g2.setFont(u8g2_font_5x7_tr);
    if (countdown) u8g2.drawStr(123, 31, String(countdown).c_str());
    else u8g2.drawStr(123, 31, "a"); // means active
}

IRAM_ATTR void loop_u8g2_u8g2ani(){ // "U8g2" animation
    u8g2.setFont(u8g2_font_8x13B_mf);
    IRAM_ATTR static int x=-8, y=8;
    u8g2.drawStr(dpox+(x+=y), dpoy+13, "U8g2");
    if (x>=127-32) y=-8;
    else if (x<=0) y=8;
}

IRAM_ATTR void loop_u8g2_speed(){ // show the speed value
    static String pr="speed";
    if (http_got_str!="" && pr!=http_got_str){
        pr=http_got_str;
    }
    if (pr!=""){
        u8g2.setFont(u8g2_font_inr27_t_cyrillic);
        u8g2.drawStr(dpox+15, dpoy+27, pr.c_str());
    }
}

IRAM_ATTR void loop_u8g2_joystick(){ // output the joystick values
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+0, dpoy+26, (String(BVOLTAGE(joystick[3]))+" "+joystick_tri[1]+" "+joystick_tri[0]+" "+joystick_tri[2]).c_str());
}

/*IRAM_ATTR */void loop_u8g2_multicast(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+0, dpoy+13, "Multicast:");
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawStr(dpox+1, dpoy+20, multicast_found.c_str());
}

/*IRAM_ATTR */void loop_u8g2_motor(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+0, dpoy+13, "Motor Control");
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawStr(dpox+1, dpoy+20, "to:");
    u8g2.drawStr(dpox+20, dpoy+20, http_request_ip);
}

/*IRAM_ATTR */void loop_u8g2_motor_cmd_list(){
    u8g2.setFont(u8g2_font_5x7_tr);
    u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, 9);
    u8g2.setFontMode(0);
    u8g2.setDrawColor(0);
    u8g2.drawStr(dpox+1-jtk_pos_x, dpoy+8, http_commands_str[jtk_pos_y]);
    u8g2.setDrawColor(1);
    u8g2.setFontMode(1);
    u8g2.drawStr(dpox+1-jtk_pos_x, dpoy+15, http_commands_str[jtk_pos_y+1]);
    u8g2.drawStr(dpox+1-jtk_pos_x, dpoy+22, http_commands_str[jtk_pos_y+2]);
    u8g2.drawStr(dpox+1-jtk_pos_x, dpoy+29, http_commands_str[jtk_pos_y+3]);
}

/*IRAM_ATTR */void loop_u8g2_motor_http_returned(){ // skip k lines to start drawing 3 lines.
    u8g2.setFont(u8g2_font_8x13B_mf);
    int y=0, z=3, i=-2, j=0, k=jtk_pos_y;
    for (; z && ((i=copy_http_got_str.indexOf('\n', j))!=-1); j=i+1){
        if (k){k--; continue;}
        u8g2.drawStr(dpox+0-jtk_pos_x, dpoy+13+y, ("["+copy_http_got_str.substring(j, i)+"]").c_str());
        y+=13; z--;
    }
}

/*IRAM_ATTR */void loop_u8g2_context(){ // skip k lines to start drawing 3 lines.
    u8g2.setFont(u8g2_font_5x7_tr);
    int y=0, z=3, i=-2, j=0, k=jtk_pos_y;
    for (; z && ((i=copy_http_got_str.indexOf('\n', j))!=-1); j=i+1){
        if (k){k--; continue;}
        u8g2.drawStr(dpox+0-jtk_pos_x, dpoy+13+y, copy_http_got_str.substring(j, i).c_str());
        y+=13; z--;
    }
}

/*IRAM_ATTR */void loop_u8g2_input_history(){ // skip k lines to start drawing 3 lines.
    u8g2.setFont(u8g2_font_8x13B_mf);
    int y=0, z=3, i=-2, j=0, k=jtk_pos_y;
    for (; z && ((i=input_str_history.indexOf('\n', j))!=-1); j=i+1){
        if (k){k--; continue;}
        if (z==3){
            u8g2.drawBox(dpox+0, dpoy+0, U8G2_X, 15);
            u8g2.setFontMode(0);
            u8g2.setDrawColor(0);
            u8g2.drawStr(dpox+0-jtk_pos_x, dpoy+13+y, ("["+input_str_history.substring(j, i)+"]").c_str());
            u8g2.setDrawColor(1);
            u8g2.setFontMode(1);
            input_str_history_posi=j;
            input_str_history_posj=i;
        }
        else u8g2.drawStr(dpox+0-jtk_pos_x, dpoy+13+y, ("["+input_str_history.substring(j, i)+"]").c_str());
        y+=13; z--;
    }
}

/*IRAM_ATTR */void loop_u8g2_local_cmds(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+0, dpoy+13, " Local Commands");
}

/*IRAM_ATTR */void loop_u8g2_remote_fcmds(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+0, dpoy+13, "Remote Fast Cmds");
}

IRAM_ATTR void loop_u8g2_connect(){
    u8g2.setFont(u8g2_font_8x13B_mf);
    u8g2.drawStr(dpox+24, dpoy+13, "connection");
    u8g2.setFont(u8g2_font_5x7_tr);
#if STA_MODE
    u8g2.drawStr(dpox+2, dpoy+22,\
        is_station_mode_established?\
        (String("-> connected: ")+netstr_ssid).c_str():\
        "-> disconnected."\
    );
    u8g2.drawStr(dpox+0, dpoy+29, "tap to connect/disconnect");
#else // AP_MODE
    u8g2.drawStr(dpox+2, dpoy+22, (String("AP mode: ")+WIFI_CRED[0][0]).c_str());
#endif // STA_MODE
}

IRAM_ATTR void loop_u8g2_network(){
    u8g2.setFont(u8g2_font_5x7_tr);
#if STA_MODE
    u8g2.drawStr(dpox+1, dpoy+8,\
        is_station_mode_established?\
        (String("Local IP: ")+netstr_localip).c_str():\
        "Local IP:"\
    );
    u8g2.drawStr(dpox+1, dpoy+15,\
        is_station_mode_established?\
        (String("AP IP: ")+netstr_remoteip).c_str():\
        "AP IP:"\
    );
    u8g2.drawStr(dpox+1, dpoy+22,\
        is_station_mode_established?\
        (String("Website: http://")+my_hostname+".local/").c_str():\
        "Website:"\
    );
    u8g2.drawStr(dpox+1, dpoy+29,\
        is_station_mode_established?\
        ("mngcenter: "+netstr_mngcenter_ip+" Link: "+web_link).c_str():\
        "mngcenter:"\
    );
#else // AP_MODE
    u8g2.drawStr(dpox+1, dpoy+8, ("Local IP: "+WiFi.softAPIP().toString()).c_str());
#endif // STA_MODE
}

/*IRAM_ATTR */void loop_u8g2_network_cont(){
    #if STA_MODE
        u8g2.setFont(u8g2_font_5x7_tr);
        u8g2.drawStr(dpox-U8G2_X, dpoy+8,\
            is_station_mode_established?\
            (String("Local IP: ")+netstr_localip).c_str():\
            "Local IP:"\
        );
        u8g2.drawStr(dpox-U8G2_X, dpoy+15,\
            is_station_mode_established?\
            (String("AP IP: ")+netstr_remoteip).c_str():\
            "AP IP:"\
        );
        u8g2.drawStr(dpox-U8G2_X, dpoy+22,\
            is_station_mode_established?\
            (String("Website: http://")+my_hostname+".local/").c_str():\
            "Website:"\
        );
        u8g2.drawStr(dpox-U8G2_X, dpoy+29,\
            is_station_mode_established?\
            ("mngcenter: "+netstr_mngcenter_ip+" Link: "+web_link).c_str():\
            "mngcenter:"\
        );
    #endif // STA_MODE
}

inline void page_u8g2_subscript(){
    DrawConfig(draw_page_2, loop_u8g2_status_051);
    DrawConfig(draw_page_2, loop_u8g2_status_052);
    DrawConfig(draw_page_2, loop_u8g2_status_053);
    DrawConfig(draw_page_2, loop_u8g2_status_054);
    DrawConfig(draw_page_2, loop_u8g2_status_055);
}

inline void page_u8g2_mainmenu(){
    DrawConfig(draw_page_2, loop_u8g2_u8g2ani);
    DrawConfig(draw_page_2, loop_u8g2_joystick);
    page_u8g2_subscript();
}

inline void page_u8g2_local_cmds(){
    DrawConfig(draw_page_2, loop_u8g2_local_cmds);
    page_u8g2_subscript();
}

inline void page_u8g2_remote_fcmds(){
    DrawConfig(draw_page_2, loop_u8g2_remote_fcmds);
}

inline void page_u8g2_multicast(){
    DrawConfig(draw_page_2, loop_u8g2_multicast);
    page_u8g2_subscript();
}

inline void page_u8g2_motor(){
    DrawConfig(draw_page_2, loop_u8g2_motor);
    page_u8g2_subscript();
}

inline void page_u8g2_motor_http_returned(){
    DrawConfig(draw_page_2, loop_u8g2_motor_http_returned);
    page_u8g2_subscript();
}

inline void page_u8g2_speed(){
    DrawConfig(draw_page_2, loop_u8g2_speed);
    page_u8g2_subscript();
}

inline void page_u8g2_input(){
    DrawConfig(draw_page_2, loop_u8g2_alphabet);
}

inline void page_u8g2_connect(){
    DrawConfig(draw_page_2, loop_u8g2_connect);
}

inline void page_u8g2_network(){
    DrawConfig(draw_page_2, loop_u8g2_network);
}

inline void page_u8g2_network_cont(){
    DrawConfig(draw_page_2, loop_u8g2_network_cont);
}

inline void page_u8g2_settings(){
    DrawConfig(draw_page_2, loop_u8g2_settings);
}

inline void page_u8g2_settings_poweronconnect(){
    DrawConfig(draw_page_2, loop_u8g2_settings_poweronconnect);
}

inline void page_u8g2_settings_poweronspeed(){ // [[main controls]]
    DrawConfig(draw_page_2, loop_u8g2_settings_poweronspeed);
}

inline void page_u8g2_settings_keepalive(){
    DrawConfig(draw_page_2, loop_u8g2_settings_keepalive);
}

inline void page_u8g2_settings_TOff1M(){
    DrawConfig(draw_page_2, loop_u8g2_settings_TOff1M);
}

inline void page_u8g2_poweroff(){
    DrawConfig(draw_page_2, loop_u8g2_poweroff);
}

void (*draw_map_x[])()={
    page_u8g2_mainmenu,         // 0
    page_u8g2_multicast,        // 1
    page_u8g2_speed,            // 2
    page_u8g2_input,            // 3
    page_u8g2_connect,          // 4
    page_u8g2_network,          // 5
    page_u8g2_network_cont,     // 6
    page_u8g2_settings,         // 7
    page_u8g2_poweroff          // 8
};

void (*draw_map_y0[])()={ // main menu base, draw_map_y_sz=3. (if mod-tbl, size should be changed as well: draw_map_y_sz)
    page_u8g2_mainmenu,
    page_u8g2_local_cmds,
    page_u8g2_remote_fcmds,
};

void (*draw_map_y1[])()={ // multicast base, draw_map_y_sz=3. (if mod-tbl, size should be changed as well: draw_map_y_sz)
    page_u8g2_multicast,
    page_u8g2_motor,
    page_u8g2_motor_http_returned,
};

void (*draw_map_y7[])()={ // settings, draw_map_y_sz=5. (if mod-tbl, size should be changed as well: draw_map_y_sz)
    page_u8g2_settings,
    page_u8g2_settings_poweronconnect,
    page_u8g2_settings_poweronspeed, // [[main controls]]
    page_u8g2_settings_keepalive,
    page_u8g2_settings_TOff1M,
};

void check_and_process_horizontal_pages(unsigned dir){ // horizontal animation allocate.
    if (dir==JS_DIR_RIGHT){ // right
        if (++mapx_idx>=sizeof(draw_map_x)/sizeof(*draw_map_x)) mapx_idx=0;
        slide_dir=SLIDE_DIR_TIPRIGHT;
        slide_idx=0;
        draw_map_x[mapx_idx]();
        slide_2_pos=U8G2_X; // the order is of concern
    }
    else { // JS_DIR_LEFT, left
        if (--mapx_idx<0) mapx_idx=sizeof(draw_map_x)/sizeof(*draw_map_x)-1;
        slide_dir=SLIDE_DIR_TIPLEFT;
        slide_idx=0;
        draw_map_x[mapx_idx]();
        slide_2_pos=-U8G2_X; // the order is of concern
    }
}

bool check_and_process_vertical_pages(unsigned dir){ // vertical animation allocate.
    void (*(*the_v_pages))()=0;
    unsigned draw_map_y_sz=0;
    switch (mapx_idx){ // the pages which have vertical pages.
        case 0:
            the_v_pages=draw_map_y0;
            draw_map_y_sz=3;
            break;
        case 1:
            the_v_pages=draw_map_y1;
            draw_map_y_sz=3;
            break;
        case 7:
            the_v_pages=draw_map_y7;
            draw_map_y_sz=5;
            break;
        default: return false;
    }
    if (dir==JS_DIR_UP){ // up
        if (--mapy_idx<0) mapy_idx=draw_map_y_sz-1;
        slide_dir=SLIDE_DIR_TIPUP;
        slide_idx=0;
        the_v_pages[mapy_idx]();
        slide_2_pos=-U8G2_Y; // the order is of concern
    }
    else { // JS_DIR_DOWN, down
        if (++mapy_idx>=draw_map_y_sz) mapy_idx=0;
        slide_dir=SLIDE_DIR_TIPDOWN;
        slide_idx=0;
        the_v_pages[mapy_idx]();
        slide_2_pos=U8G2_Y; // the order is of concern
    }
    return true;
}

void InputStrProc_RemoveFile(){
    if (cDataBase(("/"+input_str).c_str()).Remove())
        DrawHoldOnEx_MsgBoxShowing("  remove done");
    else DrawHoldOnEx_MsgBoxShowing(" file not found");
}

void InputStrProc_SetTarget(){
    strcpy(http_request_ip, input_str.c_str());
    DrawHoldOnEx_MsgBoxShowing("   done set");
}

void InputStrProcessing(void (*fptr)()=0, unsigned org_x=0, unsigned org_y=0){ // input_str and locked_lr are used.
    static void (*cb)()=0;
    static unsigned x=0, y=0;
    if (cb){
        if (locked_lr) return;
        mapx_idx=x;
        mapy_idx=y;
        jtk_pos_x=jtk_pos_y=0;
        DrawDeConfig(draw_page_1); // input window
        DrawDeConfig(draw_page_1); // input window
        DrawDeConfig(draw_page_1); // background black window
        if (input_str!="") cb();
        else DrawHoldOnEx_MsgBoxShowing(" input cancelled");
        cb=0;
    }
    else if (fptr){ // allocate and jump to input windows.
        cb=fptr;
        x=org_x;
        y=org_y;
        input_str="";
        // note, some existing windows can not be overlapped.
        DrawConfig(draw_page_1, loop_u8g2_bgblack);
        DrawConfig(draw_page_1, loop_u8g2_alphabet);
        DrawConfig(draw_page_1, loop_u8g2_alphabet_change);
        mapx_idx=3; mapy_idx=0; locked_lr=1; // switch to input page state. note that switch back to only has locked_lr=0.
        jtk_pos_x=5; jtk_pos_y=2; // locate cursor at "S"
    }
}

void joystick_btn_functions(){ // joystick button response function.
    static unsigned to_upper=0; // character case. (0, lower); (1, upper once); (2, upper fixed).
    switch (mapx_idx){
        case 0: // main menu base
            switch (mapy_idx){
                case 0: // the main menu base
                    break;
                case 1: // local cmds
                    if (!locked_lr){
                        locked_lr=1;
                        jtk_pos_x=0;
                        jtk_pos_y=0;
                        opt_i_box[0]="     list directory";
                        opt_i_box[1]="       remove file";
                        opt_i_box[2]="    firmware version";
                        opt_i_box[3]="         cancel";
                        DrawConfig(draw_page_1, loop_u8g2_bgblack);
                        DrawConfig(draw_page_1, loop_u8g2_options_4_boxes);
                    }
                    else if (!locked_ud){ // ls, rm, version; use locked_ud to processing elsewhere.
                        switch (jtk_pos_y){
                            case 0: // ls
                                copy_http_got_str="";
                                cDataBase::Ls("/", &copy_http_got_str);
                                copy_http_got_str.replace("\r\n", "\n"); // rebuild string for drawing.
                                // note, can use overlap however loop_u8g2_options_4_boxes can not be.
                                DrawDeConfig(draw_page_1);
                                DrawConfig(draw_page_1, loop_u8g2_context);
                                locked_ud=1;
                                return; // must return.
                            case 1: // rm
                                // note, can use overlap however loop_u8g2_options_4_boxes can not be.
                                DrawDeConfig(draw_page_1); // loop_u8g2_options_4_boxes
                                DrawDeConfig(draw_page_1); // loop_u8g2_bgblack
                                // so need more treatments if sibling also has input.
                                DrawHoldOnEx_MsgBoxShowing(" input filename");
                                locked_lr=0; // it is must have and before InputStrProcessing.
                                InputStrProcessing(InputStrProc_RemoveFile, mapx_idx, mapy_idx);
                                break;
                            case 2: // version
                                jtk_pos_y=0;
                                copy_http_got_str=String(theVERSION)+"\n";
                                DrawDeConfig(draw_page_1);
                                DrawConfig(draw_page_1, loop_u8g2_context);
                                locked_ud=1;
                                return; // must return.
                            case 3: // cancel
                                DrawDeConfig(draw_page_1); // loop_u8g2_options_4_boxes
                                DrawDeConfig(draw_page_1); // loop_u8g2_bgblack
                                locked_lr=0;
                                break;
                        }
                    }
                    break;
                case 2: // remote fast cmds
                    if (!locked_lr){
                        locked_lr=1;
                        jtk_pos_x=0;
                        jtk_pos_y=0;
                        DrawConfig(draw_page_1, loop_u8g2_bgblack);
                        DrawConfig(draw_page_1, loop_u8g2_options_16_boxes_prepare);
                        DrawConfig(draw_page_1, loop_u8g2_options_16_boxes);
                    }
                    else if ((jtk_pos_x==0) && (jtk_pos_y==0)){
                        DrawDeConfig(draw_page_1); // loop_u8g2_options_16_boxes
                        DrawDeConfig(draw_page_1); // loop_u8g2_options_16_boxes_prepare
                        DrawDeConfig(draw_page_1); // loop_u8g2_bgblack
                        locked_lr=0;
                    }
                    else if (*http_request_ip){
                        String a=String("http://")+http_request_ip+remote_fast_commands[jtk_pos_y][jtk_pos_x];
                        if (HTTPC_http_client_connect(a.c_str())){
                            copy_http_got_str=http_got_str;
                            DrawHoldOnEx_MsgBoxShowing("       O");
                        }
                        else DrawHoldOnEx_MsgBoxShowing("        X");
                    }
                    break;
            }
            break;
        case 1: // multicast base
            switch (mapy_idx){
                case 0: // the multicast base
                    #if STA_MODE
                    if (!locked_lr){
                        locked_lr=1;
                        jtk_pos_x=multicast_is_broadcasting;
                        jtk_pos_y=0;
                        opt_i_box[0]="Disable"; opt_i_box[1]="Transmit"; opt_i_box[2]="Receive";
                        DrawEnd(draw_page_1);
                        DrawConfig(draw_page_1, loop_u8g2_options_3_boxes);
                    }
                    else { // control multicast off/tr/rv, in addition, do the re-request.
                        multicast_tr_count="-";
                        multicast_found="";
                        extern void setup_select_multicast(int);
                        ///// setup_select_multicast(jtk_pos_x+3); // +3 means need to re-request.
                        setup_select_multicast(jtk_pos_x); /////
                        DrawEnd(draw_page_1);
                        DrawConfig(draw_page_1, loop_u8g2_multicast);
                        locked_lr=0;
                    }
                    #endif // STA_MODE
                    break;
                case 1: // motor
                    if (!locked_lr){
                        locked_lr=1;
                        jtk_pos_x=0;
                        jtk_pos_y=0;
                        DrawEnd(draw_page_1);
                        DrawConfig(draw_page_1, loop_u8g2_motor_cmd_list);
                    }
                    else if (!locked_ud){ // send command and wait for response. locked_ud later added in order dispatch to elsewhere.
                        if (jtk_pos_y==sizeof(http_commands_str)/sizeof(http_commands_str[0])-5){ // not the cmd but the selecting device.
                            if (!motor_servicedev_query()){ // to select a motor device
                                DrawHoldOnEx_MsgBoxShowing("Target not found"); // not found, use default(AP).
                                return;
                            }
                            else { // it is going to create another message box for selection.
                                DrawConfig(draw_page_1, loop_u8g2_bgblack);
                                DrawConfig(draw_page_1, loop_u8g2_options_4_boxes);
                                jtk_pos_x=0;
                                jtk_pos_y=0;
                                locked_ud=1; // need to release locked_ud and locked_lr when later done.
                                return; // here must return rather than break.
                            }
                        }
                        else if (jtk_pos_y==0){ // "ANY URL"
                            if (!motor_cmds_selection(jtk_pos_y)) return;
                            motor_http_request_passthrough=1;
                        }
                        else if (jtk_pos_y==1){ // "SET TARGET"
                            DrawEnd(draw_page_1);
                            DrawConfig(draw_page_1, loop_u8g2_motor);
                            locked_lr=0; // it is must have and before InputStrProcessing.
                            InputStrProcessing(InputStrProc_SetTarget, mapx_idx, mapy_idx);
                            DrawHoldOnEx_MsgBoxShowing(" input target ip");
                            break;
                        }
                        else if ((jtk_pos_y!=sizeof(http_commands_str)/sizeof(http_commands_str[0])-4) && !motor_cmds_selection(jtk_pos_y))
                            return; // not cancel and command selected failed. another failure would be network not available.
                        locked_lr=0;
                        DrawEnd(draw_page_1);
                        DrawConfig(draw_page_1, loop_u8g2_motor);
                    }
                    break;
                case 2: // http returned string
                    if (!locked_lr){
                        locked_lr=1;
                        jtk_pos_x=0;
                        jtk_pos_y=0;
                        DrawEnd(draw_page_1);
                        DrawConfig(draw_page_1, loop_u8g2_motor_http_returned);
                    }
                    else {
                        jtk_pos_x=0;
                        jtk_pos_y=0;
                        locked_lr=0;
                    }
                    break;
            }
            break;
        case 2: // speed
            switch (mapy_idx){
                case 0: // speed. pause it or not. only for the auto http request behavior.
                    if (only_pause_the_auto_http_req){
                        only_pause_the_auto_http_req=false;
                        DrawHoldOnEx_MsgBoxShowing("     paused");
                        slowdown=0;
                    }
                    else {
                        only_pause_the_auto_http_req=true;
                        DrawHoldOnEx_MsgBoxShowing("    continue");
                        slowdown=1000;
                    }
                    break;
            }
            break;
        case 3: // input
            if (!locked_lr){
                locked_lr=1;
                jtk_pos_x=0;
                jtk_pos_y=0;
                //// to_upper=0;
                input_str="";
                DrawConfig(draw_page_1, loop_u8g2_alphabet_change);
            }
            else if (input_str_history_flag){ // treat with the history mode escape overhere. note that order is of concern.
                if (input_str_history_posi){ // if zero means the 1st string which must be "Cancel".
                    if (input_str_history_flag==1) input_str=input_str_history.substring(input_str_history_posi, input_str_history_posj);
                    else if (input_str_history_flag==2) input_str+=input_str_history.substring(input_str_history_posi, input_str_history_posj);
                    else { // history insert mode
                        input_str_history_flag-=3;
                        input_str=\
                            input_str.substring(0, input_str_history_flag-1)+\
                            input_str_history.substring(input_str_history_posi, input_str_history_posj)+\
                            input_str.substring(input_str_history_flag-1);
                        input_str_insmode+=\
                            input_str_history.substring(input_str_history_posi, input_str_history_posj).length(); // remember to update.
                        input_str_history_flag=0;
                    }
                    jtk_pos_x=input_str_history_flag; jtk_pos_y=3;
                }
                input_str_history_flag=0;
                input_str_history="";
                DrawDeConfig(draw_page_1); // DrawConfig(draw_page_1, loop_u8g2_input_history).
                DrawDeConfig(draw_page_1); // DrawConfig(draw_page_1, loop_u8g2_bgblack).
            }
            else if ((jtk_pos_x==5) && (jtk_pos_y==2)){ // save and exit
                unsigned the_len=input_str.length();
                if (the_len && (input_str.indexOf('\r')==-1)){
                    input_str_history_notsave=input_str+"\n"+input_str_history_notsave;
                    input_str_history=input_str+"\n"+input_str_history;
                }
                input_str_insmode=0;
                locked_lr=0;
            }
            else if ((jtk_pos_x==4) && (jtk_pos_y==2)){ // backspace
                if (input_str_insmode){
                    if (input_str_insmode>1){
                        input_str_insmode--;
                        input_str=input_str.substring(0, input_str_insmode-1)+input_str.substring(input_str_insmode);
                    }
                }
                else if (input_str.length()) input_str.remove(input_str.length()-1);
            }
            else if ((jtk_pos_x==3) && (jtk_pos_y==2)){ // clear all
                input_str_insmode=0;
                input_str="";
            }
            else if ((jtk_pos_x==2) && (jtk_pos_y==2)){ // to upper fixed
                if (to_upper!=2) to_upper=2;
                else to_upper=0;
            }
            else if ((jtk_pos_x==1) && (jtk_pos_y==2)){ // to upper once
                to_upper=1;
            }
            else if ((jtk_pos_x==0) && (jtk_pos_y==2)){ // line-feed
                if (input_str_insmode){
                    input_str=input_str.substring(0, input_str_insmode-1)+"\r\n"+input_str.substring(input_str_insmode-1);
                    input_str_insmode+=2;
                }
                else input_str+="\r\n";
            }
            else if (jtk_pos_y==3){ // enhance mode
                if (jtk_pos_x==0){ // insmode
                    if (!input_str_insmode_flag){ // not yet enter insert mode
                        if (!input_str_insmode && input_str.length()){ // pass and entering insert mode
                            jtk_pos_x=input_str_insmode=input_str.length(); // set the initial offset of x
                            input_str_insmode_flag=1; // in order to do cursor positioning
                        }
                        else input_str_insmode=0; // escape insert mode. input_str_insmode and flag are mutex.
                    }
                }
                else if (input_str_insmode_flag){input_str_insmode_flag=0; input_str_insmode=jtk_pos_x;} // escape insert mode positioning
                else if (!input_str_insmode || (jtk_pos_x==1) || (jtk_pos_x==2)){ // the other 4 modes.
                    // and moreover let history replace/append mode pass through for inserting history.
                    switch (jtk_pos_x){
                        case 0: break; // insert mode, already impl.
                        case 1: // history replace/1 or append/2
                        case 2: // history replace/1 or append/2
                            input_str_history_flag=jtk_pos_x;
                            if (input_str_insmode) input_str_history_flag=input_str_insmode+3; // for history insert mode.
                            extern void InputLoadHistory();
                            InputLoadHistory();
                            DrawConfig(draw_page_1, loop_u8g2_bgblack);
                            DrawConfig(draw_page_1, loop_u8g2_input_history);
                            jtk_pos_x=jtk_pos_y=0;
                            break;
                        case 3: // save history mode
                            extern bool InputSaveHistory();
                            if (InputSaveHistory()) DrawHoldOnEx_MsgBoxShowing("saved to history");
                            else DrawHoldOnEx_MsgBoxShowing(" nothing to save");
                            break;
                        case 4: break; // keep
                    }
                }
            }
            else { // user input
                if (input_str_insmode){
                    if (to_upper==1){
                        String a(alphabet[jtk_pos_y][jtk_pos_x]);
                        a.toUpperCase();
                        input_str=input_str.substring(0, input_str_insmode-1)+a+input_str.substring(input_str_insmode-1);
                        to_upper=0;
                    }
                    else if (to_upper==2){
                        String a(alphabet[jtk_pos_y][jtk_pos_x]);
                        a.toUpperCase();
                        input_str=input_str.substring(0, input_str_insmode-1)+a+input_str.substring(input_str_insmode-1);
                    }
                    else input_str=\
                        input_str.substring(0, input_str_insmode-1)+alphabet[jtk_pos_y][jtk_pos_x]+input_str.substring(input_str_insmode-1);
                    input_str_insmode++;
                }
                else {
                    if (to_upper==1){String a(alphabet[jtk_pos_y][jtk_pos_x]); a.toUpperCase(); input_str+=a; to_upper=0;}
                    else if (to_upper==2){String a(alphabet[jtk_pos_y][jtk_pos_x]); a.toUpperCase(); input_str+=a;}
                    else input_str+=alphabet[jtk_pos_y][jtk_pos_x];
                }
            }
            break;
        case 4: // connect
            copy_sta_do_the_connect^=1;
            break;
        case 5: // network info main window
            break;
        case 6: // network info sub window. we use it for joystick remote control in direction signal.
            if (!locked_lr){
                DrawHoldOnEx_MsgBoxShowing(" js ctrl entered");
                locked_lr=1;
                jtk_pos_x=0;
                jtk_pos_y=0;
                if (multicast_is_broadcasting) multicast_manual=1;
            }
            else {
                if (multicast_manual) speed_str="0"; // multicast version
                else { // http version
                    static unsigned retry=3;
                    if (retry){
                        if (--retry && !HTTPC_http_client_connect((String("http://")+http_request_ip+"/t").c_str())){
                            DrawHoldOnEx_MsgBoxShowing("      wait");
                            break;
                        }
                        else retry=3;
                    }
                }
                multicast_manual=0;
                DrawHoldOnEx_MsgBoxShowing("     exited");
                locked_lr=0;
            }
            break;
        case 7: // settings
            switch (mapy_idx){
                case 1: // page_u8g2_settings_poweronconnect
                    {
                        timer_draw_manip.Pause(); ///// pause the timer and the delay are all must else wdtreset.
                        delay(300); ///// pause the timer and the delay are all must else wdtreset.
                        unsigned a=0;
                        cDataBase s("/settings");
                        s.Open();
                        s.ReadWords(a, "copy_sta_do_the_connect", 0);
                        s.Close();
                        s.ReplaceWordsEx("copy_sta_do_the_connect", 0, a^=1);
                        timer_draw_manip.Resume();
                        DrawEnd(draw_page_1);
                        if (a) DrawConfig(draw_page_1, loop_u8g2_on);
                        else DrawConfig(draw_page_1, loop_u8g2_off);
                    }
                    break;
                case 2: // page_u8g2_settings_poweronspeed. [[main controls]]
                    {
                        timer_draw_manip.Pause(); ///// pause the timer and the delay are all must else wdtreset.
                        delay(300); ///// pause the timer and the delay are all must else wdtreset.
                        unsigned a=0;
                        cDataBase s("/settings");
                        s.Open();
                        s.ReadWords(a, "pwron_spd_conn", 0);
                        s.Close();
                        s.ReplaceWordsEx("pwron_spd_conn", 0, a^=1);
                        timer_draw_manip.Resume();
                        DrawEnd(draw_page_1);
                        if (a) DrawConfig(draw_page_1, loop_u8g2_on);
                        else DrawConfig(draw_page_1, loop_u8g2_off);
                    }
                    break;
                case 3: // page_u8g2_settings_keepalive
                    {
                        timer_draw_manip.Pause(); ///// pause the timer and the delay are all must else wdtreset.
                        delay(300); ///// pause the timer and the delay are all must else wdtreset.
                        cDataBase s("/settings");
                        s.ReplaceWordsEx("pwron_keep_alive", 0, pwron_keep_alive^=1);
                        timer_draw_manip.Resume();
                        DrawEnd(draw_page_1);
                        if (pwron_keep_alive) DrawConfig(draw_page_1, loop_u8g2_on);
                        else DrawConfig(draw_page_1, loop_u8g2_off);
                    }
                    break;
                case 4: // page_u8g2_settings_TOff1M
                    stop_the_timer=3;
                    break;
            }
            break;
        case 8: // poweroff
            DrawEnd(draw_page_1);
            DrawConfig(draw_page_1, loop_u8g2_bgblack);
            poweroff=1;
            break;
    }
    if (locked_ud){
        if ((mapx_idx==1) && (mapy_idx==1)){ // (motor ctrl, select device)
            DrawDeConfig(draw_page_1); // it is DrawConfig(draw_page_1, loop_u8g2_options_4_boxes)
            DrawDeConfig(draw_page_1); // it is DrawConfig(draw_page_1, loop_u8g2_bgblack)
            if ((jtk_pos_y<3) && *(opt_i_box[jtk_pos_y])){
                strcpy(http_request_ip, opt_i_box[jtk_pos_y]); // assign the tgt dev ip.
                DrawHoldOnEx_MsgBoxShowing(http_request_ip);
            }
            locked_ud=0;
            //// locked_lr=0;
        }
        else if ((mapx_idx==0) && (mapy_idx==1)){ // (main, local cmds)
            DrawDeConfig(draw_page_1); // it is DrawConfig(draw_page_1, loop_u8g2_context)
            DrawDeConfig(draw_page_1); // it is DrawConfig(draw_page_1, loop_u8g2_bgblack)
            locked_ud=0;
            locked_lr=0;
        }
    }
}

void joystick_dir_funcs(unsigned dir){ // joystick direction response function.
    // mainly control the directional behaviors usually the positional boundaries or remap when in an entered page.
    // use mapx_idx, mapy_idx to positioning.
    static unsigned tmpx=0;

    if (locked_ud){
        if ((mapx_idx==1) && (mapy_idx==1)){ // (motor ctrl, select device)
            if (jtk_pos_y>3) jtk_pos_y=0;
            else if (jtk_pos_y<0) jtk_pos_y=3;
        }
        else if ((mapx_idx==0) && (mapy_idx==1)){ // (main, local cmds)
            if (dir==JS_DIR_LEFT) jtk_pos_x-=32;
            else if (dir==JS_DIR_RIGHT) jtk_pos_x+=32;
            if (jtk_pos_x<0) jtk_pos_x=0;
            else if (jtk_pos_x>512) jtk_pos_x=0;
            if (jtk_pos_y<0) jtk_pos_y=0;
        }
    }

    else switch (mapx_idx){
        case 0: // main menu, local cmds
            switch (mapy_idx){
                case 0: // main menu
                    break;
                case 1: // local cmds
                    if (jtk_pos_y>3) jtk_pos_y=0;
                    else if (jtk_pos_y<0) jtk_pos_y=3;
                    break;
                case 2: // remote fast cmds
                    // because currently only one direction is applied at a time.
                    // that is, no slant condition.
                    if (jtk_pos_x>3){jtk_pos_x=0; if (++jtk_pos_y>3) jtk_pos_y=0;}
                    else if (jtk_pos_x<0){jtk_pos_x=3; if (--jtk_pos_y<0) jtk_pos_y=3;}
                    else if (jtk_pos_y>3){jtk_pos_y=0; if (++jtk_pos_x>3) jtk_pos_x=0;}
                    else if (jtk_pos_y<0){jtk_pos_y=3; if (--jtk_pos_x<0) jtk_pos_x=3;}
                    break;
            }
            break;
        case 1: // multicast, motor, http returned string
            switch (mapy_idx){
                case 0: // multicast
                    if (jtk_pos_x>2) jtk_pos_x=0;
                    else if (jtk_pos_x<0) jtk_pos_x=2;
                    break;
                case 1: // motor commands
                    if (jtk_pos_y>int(sizeof(http_commands_str)/sizeof(http_commands_str[0])-4)) jtk_pos_y=0; // int() is must.
                    else if (jtk_pos_y<0) jtk_pos_y=sizeof(http_commands_str)/sizeof(http_commands_str[0])-4;
                    if (dir==JS_DIR_LEFT) jtk_pos_x-=32;
                    else if (dir==JS_DIR_RIGHT) jtk_pos_x+=32;
                    if (jtk_pos_x<0) jtk_pos_x=0;
                    break;
                case 2: // http returned string
                    if (jtk_pos_y<0) jtk_pos_y=0;
                    if (dir==JS_DIR_LEFT) jtk_pos_x-=32;
                    else if (dir==JS_DIR_RIGHT) jtk_pos_x+=32;
                    if (jtk_pos_x<0) jtk_pos_x=0;
                    break;
            }
            break;
        case 3: // input
            if (input_str_history_flag){ // means currently in history mode
                if (dir==JS_DIR_LEFT) jtk_pos_x-=32;
                else if (dir==JS_DIR_RIGHT) jtk_pos_x+=32;
                if (jtk_pos_x>1022) jtk_pos_x=1022;
                else if (jtk_pos_x<0) jtk_pos_x=0;
                if (jtk_pos_y<0) jtk_pos_y=0;
            }
            else {
                if (jtk_pos_y>=3){
                    if (!tmpx) tmpx=jtk_pos_x+1; // this x would not over 24.
                    jtk_pos_y=3;

                    if (jtk_pos_x>1022) jtk_pos_x=1022; // these 2 lines still effective,
                    else if (jtk_pos_x<0) jtk_pos_x=0; // because x may over 24.

                }
                else {
                    if (jtk_pos_x>24) jtk_pos_x=24;
                    else if (jtk_pos_x<0) jtk_pos_x=0;
                    if (!input_str_insmode_flag){
                        if (tmpx){jtk_pos_x=tmpx-1; tmpx=0;}
                        if (jtk_pos_y<0) jtk_pos_y=0;
                    }
                    else jtk_pos_y=3;
                }
                if (jtk_pos_y==3){
                    if (input_str_insmode_flag){ // the insert mode. so it limited to 1022 chars priorly.
                        if (jtk_pos_x>input_str.length()) jtk_pos_x=1;
                        else if (jtk_pos_x<1) jtk_pos_x=input_str.length();
                        input_str_insmode=jtk_pos_x;
                    }
                    else if (jtk_pos_x>4) jtk_pos_x=4; // the enhance mode
                }
            }
            break;
        case 6: // network info sub window. we use it for joystick remote control in direction signal.
            if (multicast_manual){ // multicast version
                if (dir==JS_DIR_LEFT) speed_str="20000"; // close
                else if (dir==JS_DIR_RIGHT) speed_str="10000"; // open
                else if (dir==JS_DIR_UP) speed_str="30000"; // inc open
                else if (dir==JS_DIR_DOWN) speed_str="40000"; // close
            }
            else {// http version
                if (dir==JS_DIR_LEFT) HTTPC_http_client_connect((String("http://")+http_request_ip+"/t?t=2").c_str());
                else if (dir==JS_DIR_RIGHT) HTTPC_http_client_connect((String("http://")+http_request_ip+"/t?t=1").c_str());
                else if (dir==JS_DIR_UP) HTTPC_http_client_connect((String("http://")+http_request_ip+"/t?t=3").c_str());
                else if (dir==JS_DIR_DOWN) HTTPC_http_client_connect((String("http://")+http_request_ip+"/t?t=4").c_str());
            }
            break;
    }
}

bool DrawHoldOn_Setup(){ // return true if setup success.
    if (DrawHoldOn_state) return false; // need to redo later.
    DrawHoldOn_state=DRAW_HOLDON_ISSUED;
    return true;
}

bool DrawHoldOn_Answer(void (*fp)()){ // return true if submit success.
    if (!DrawHoldOn_state) return false; // because the response became outdate.
    DrawHoldOn_resp_fptr=fp;
    return true;
}

bool DrawHoldOn_Polling(){ // return true if need blocked to wait for the message.
    static unsigned tmo;
    switch (DrawHoldOn_state){
        case DRAW_HOLDON_NEGATIVE: return false;
        case DRAW_HOLDON_ISSUED:
            tmo=micros();
            DrawHoldOn_state=DRAW_HOLDON_ING;
        case DRAW_HOLDON_ING:
            if (DrawHoldOn_resp_fptr)
                DrawConfig(draw_page_1, DrawHoldOn_resp_fptr); // note that it is possible not display since removed by others.
            else if (Timeout(DRAW_HOLDON_RESP_TMO, tmo)) DrawConfig(draw_page_1, loop_u8g2_no_resp);
            else return true;
            DrawHoldOn_state=DRAW_HOLDON_RESPED;
            tmo=micros();
        case DRAW_HOLDON_RESPED:
            if (Timeout(DRAW_HOLDON_DRAW_TMO, tmo)){
                DrawDeConfig(draw_page_1); // note that it is possible yielding present page empty since only existing a layer.
                DrawHoldOn_resp_fptr=0;
                DrawHoldOn_state=DRAW_HOLDON_NEGATIVE;
                return false;
            }
            return true;
    }
}

bool DrawHoldOnEx_MsgBoxShowing(String msg){ // assign a message string then it will show up soon.
    msg.toCharArray(loop_u8g2_message_str, 17);
    if (!DrawHoldOn_Setup()) return false;
    if (!DrawHoldOn_Answer(loop_u8g2_message)) return false;
    return true;
}

bool DrawHoldOnEx_isdoneMsgBoxShowing(){ // since it would alter the current draws, disrupt if change draws on other sides.
    // so make sure done before going on(changing draws elsewhere) is better.
    return !DrawHoldOn_Polling();
}

void DrawProcessing(){ // the main drawing function where according to joystick behaviors.

    // ------ first, need to get a useful of reproduced joystick behaviors, the joystick_dir.
    if (!joystick_action){
        joystick_dir=0;
        if (joystick_x_old!=joystick_tri[1]){joystick_x_old=joystick_tri[1];joystick_dir|=((joystick_x_old&1)? 1: 2);}
        if (joystick_y_old!=joystick_tri[0]){joystick_y_old=joystick_tri[0];joystick_dir|=((joystick_y_old&1)? 4: 8);}
        static unsigned breakcode=1;
        if (joystick_b_old!=joystick_tri[2]){joystick_b_old=joystick_tri[2]; if (breakcode){joystick_dir|=16; breakcode=0;}}
        else breakcode=1;
    }
    else {
        joystick_x_old=joystick_tri[1];
        joystick_y_old=joystick_tri[0];
        joystick_b_old=joystick_tri[2];
    }

    // ------ next, control the incoming event.
    if (slide_2_pos) return;
    if (draw_page_2){ // means done the sliding draw and to deallocate.
        DrawEnd(draw_page_1);
        draw_page_1=draw_page_2;
        draw_page_2=0;
        joystick_action=0;
        DrawSubmit=0; // change from 2 to 0, deallocate draw_page_2(1), in order to reenable the refresh
    }
    DrawHoldOn_Polling(); // showing floating message
    InputStrProcessing(); // input polling.

    if (joystick_dir){
        unsigned i=1;
        do {
            if (slide_2_pos) break;
            switch (joystick_dir&i){ // jtk_pos_x, jtk_pos_y might have sync problem when change over here.
                case JS_DIR_IDLE: break;
                case JS_DIR_LEFT: // left
                    if (locked_lr){
                        jtk_pos_x--;
                        joystick_dir_funcs(JS_DIR_LEFT);
                        joystick_action=0;
                        joystick_dir=0;
                        return;
                    }
                    mapy_idx=0; // y is submenu, so needs to reset.
                    check_and_process_horizontal_pages(JS_DIR_LEFT);
                    break;
                case JS_DIR_RIGHT: // right
                    if (locked_lr){
                        jtk_pos_x++;
                        joystick_dir_funcs(JS_DIR_RIGHT);
                        joystick_action=0;
                        joystick_dir=0;
                        return;
                    }
                    mapy_idx=0; // y is submenu, so needs to reset.
                    check_and_process_horizontal_pages(JS_DIR_RIGHT);
                    break;
                case JS_DIR_DOWN: // down
                    if (locked_lr){
                        jtk_pos_y++;
                        joystick_dir_funcs(JS_DIR_DOWN);
                        joystick_action=0;
                        joystick_dir=0;
                        return;
                    }
                    else if (!check_and_process_vertical_pages(JS_DIR_DOWN)){
                        joystick_action=0;
                        joystick_dir=0;
                        return;
                    }
                    break;
                case JS_DIR_UP: // up
                    if (locked_lr){
                        jtk_pos_y--;
                        joystick_dir_funcs(JS_DIR_UP);
                        joystick_action=0;
                        joystick_dir=0;
                        return;
                    }
                    else if (!check_and_process_vertical_pages(JS_DIR_UP)){
                        joystick_action=0;
                        joystick_dir=0;
                        return;
                    }
                    break;
                case JS_DIR_PUSH: // push
                    joystick_btn_functions();
                    joystick_action=0;
                    joystick_dir=0;
                    return;
                default: break;
            }
        }
        while ((i<<=1)<=JS_DIR_PUSH); // 16

        joystick_action=1;
        joystick_dir=0;
    }
}


// ------------------------------------ http processing and multicast tr.
void my_http_request(){ // called only if http_request_ip assigned.
    static unsigned tmo=micros();
    if (Timeout(266+slowdown, tmo)){ // each 0.266 seconds to send http request.

        #if STA_MODE
        if (multicast_is_broadcasting==MULTICAST_RV){ // the current revised version will no longer auto switch to http if mc recved none.
            static unsigned mc_rv_alive_state=0, tmo1=0;
            if (receiveMulticast()){http_got_str=multicast_got_or_put_str; mc_rv_alive_state=1;}
            else if (mc_rv_alive_state==1){++mc_rv_alive_state; tmo1=micros();}
            else if ((mc_rv_alive_state==2) && Timeout(15000, tmo1)){ // after 15 seconds to retry to renew.
                if (HTTPC_http_client_connect(netstr_mngcenter_web.c_str())){ // to get the if new ip.
                    netstr_mngcenter_ip=http_got_str;
                    strcpy(http_request_ip, http_got_str.c_str());
                    mc_rv_alive_state=0;
                }
                else ++mc_rv_alive_state;
            }
            else if ((mc_rv_alive_state==3) && Timeout(45000, tmo1)){ // after 45 seconds to reconnect AP
                sta_do_the_connect=0; // reconnecting
                mc_rv_alive_state=0;
            }
        }
        else 
        #endif // STA_MODE

        if (only_pause_the_auto_http_req) HTTPC_http_client_connect(web_link.c_str());

        // put the multicast transmitter here. it is not proper here but meets timing.
        #if STA_MODE
        if (multicast_is_broadcasting==MULTICAST_TR){
            if (multicast_manual){
                multicast_got_or_put_str=speed_str;
                speed_str=50000;
            }
            else {
                my_calculation_s();
                multicast_got_or_put_str=speed_str;
            }
            transmitMulticast();
        }
        #endif // STA_MODE

    }

    // http motor commands processing
    motor_cmds_processing();
}

bool motor_servicedev_query(){ // query by the identifier (motor, tcp). results assigned to option box for select. limited to 3.
    if (int n=mDnsQueryStart("motor", "tcp")){
        motor_devs_ips[0][0]=motor_devs_ips[1][0]=motor_devs_ips[2][0]=0;
        if (n>3) n=3;
        int i=0;
        for (; i<n; i++){
            IPAddress ip;
            unsigned port;
            String hname;
            mDnsQueryCont(i, ip, port, hname);
            strcpy(motor_devs_ips[i], ip.toString().c_str());
            opt_i_box[i]=motor_devs_ips[i];
        }
        mDnsQueryEnd();
        for (; i<3; i++) opt_i_box[i]=motor_devs_ips[i];
        opt_i_box[3]="Cancel";
        return true;
    }
    return false;
}

bool motor_cmds_selection(unsigned cmd){
    if (motor_issued_cmd_state) return false; // still has http processing.
    if (cmd<sizeof(http_commands_need_params)/sizeof(http_commands_need_params[0])){ // need input.
        if (http_commands_need_params[cmd]==1) motor_issued_cmd_state=1; // need 1 params.
        else if (http_commands_need_params[cmd]==2) motor_issued_cmd_state=2; // need 2 params.
    }
    else motor_issued_cmd_state=3;
    motor_issued_cmd_str=http_commands_str[cmd];
    return true;
}

void motor_cmds_processing(){
    static int x, y;
    static unsigned n;

    switch (motor_issued_cmd_state){
        case 0: return;
        case 1: // shows a message of 1 param needed.
            if (!DrawHoldOnEx_MsgBoxShowing("input 1 param")) return;
            n=1;
            motor_issued_cmd_state=4;
            break;
        case 2: // shows a message of 2 params needed.
            if (!DrawHoldOnEx_MsgBoxShowing("input 2 params")) return;
            n=3;
            motor_issued_cmd_state=4;
            break;
        case 3:
            x=mapx_idx; y=mapy_idx; // keep the present page state.
            motor_issued_cmd_state=7;
            break;
        case 4:
            if (!DrawHoldOnEx_isdoneMsgBoxShowing()) return;
            input_str=input_param[0]=input_param[1]="";
            DrawEnd(draw_page_1);
            DrawConfig(draw_page_1, loop_u8g2_alphabet);
            DrawConfig(draw_page_1, loop_u8g2_alphabet_change);
            x=mapx_idx; y=mapy_idx; // keep the present page state.
            jtk_pos_x=5; jtk_pos_y=2; // locate cursor at "S"
            mapx_idx=3; mapy_idx=0; locked_lr=1; // switch to input page state. note that switch back to only has locked_lr=0.
            motor_issued_cmd_state=5;
            break;
        case 5: // note that when done the input of 1st one, case 5 may not entered before user altered display where problematic.
            if (!locked_lr){ // means done the input.
                if (input_str==""){ // means abort.
                    DrawHoldOnEx_MsgBoxShowing("input aborted");
                    motor_issued_cmd_state=10;
                    break;
                }
                else {
                    switch (n){
                        case 1: // done 1 param
                            input_param[0]=input_str;
                            motor_issued_cmd_state=6;
                            break;
                        case 2: // done 2nd param
                            input_param[1]=input_str;
                            motor_issued_cmd_state=6;
                            break;
                        case 3: // done 1st param
                            input_param[0]=input_str;
                            input_str="";
                            jtk_pos_x=5; jtk_pos_y=2; // locate cursor at "S"
                            n=2; mapx_idx=3; mapy_idx=0; locked_lr=1;
                            return;
                    }
                    input_str="";
                }
            }
            else break;
        case 6: // process strings.
            for (int i=0, j=0, k=0; i<n; i++){
                j=motor_issued_cmd_str.indexOf('<');
                k=motor_issued_cmd_str.indexOf('>', j);
                if (j!=-1 && k!=-1){
                    input_str=motor_issued_cmd_str.substring(k+1);
                    motor_issued_cmd_str.remove(j);
                    motor_issued_cmd_str.concat(input_param[i]);
                    motor_issued_cmd_str.concat(input_str);
                }
                else { // unexpected.
                    DrawHoldOnEx_MsgBoxShowing("unexpected");
                    motor_issued_cmd_state=10;
                    return;
                }
            }
        case 7: // finalize command.
            {
                int l=motor_issued_cmd_str.indexOf(';');
                if (l!=-1) motor_issued_cmd_str.remove(l);
            }
            motor_issued_cmd_state=8;
        case 8:
            if (!DrawHoldOn_Setup()) return; // the previous hold-on is still processing.
        case 9:
            if (motor_http_request_passthrough){ // "ANY URL"
                motor_http_request_passthrough=0;
                if (HTTPC_http_client_connect(motor_issued_cmd_str.c_str())) DrawHoldOn_Answer(loop_u8g2_success);
                else DrawHoldOn_Answer(loop_u8g2_failure);
            }
            else if (HTTPC_http_client_connect(("http://"+String(http_request_ip)+motor_issued_cmd_str).c_str()))
                DrawHoldOn_Answer(loop_u8g2_success);
            else DrawHoldOn_Answer(loop_u8g2_failure);
            copy_http_got_str=http_got_str;
            // rebuild string for drawing.
            copy_http_got_str.replace("\r\n", "\n");
            copy_http_got_str.replace("\r", "\n");
            copy_http_got_str.trim();
            copy_http_got_str+="<--\n<--EOF\n";
            copy_http_got_str.replace("\n\n", "\n");
            /// printf("\r\nmotor command[http://%s%s]\r\n[%s]\r\n", http_request_ip, motor_issued_cmd_str.c_str(), http_got_str.c_str());
        case 10:
            DrawEnd(draw_page_1);
            DrawConfig(draw_page_1, loop_u8g2_motor);
            mapx_idx=x;
            mapy_idx=y;
            locked_lr=0;
            motor_issued_cmd_state=0;
    }
}

void my_calculation_s(){
    int i=(random(3)? -1: 1);
    static int j=0;
    static unsigned k=0;
    if (!k && random(17)>14) k=2;
    else if (k){k--; i=1;}
    j+=i;
    if (j<0) j=0;
    else if (j>7) j=7;
    speed_str=(j*1000+random(1000));
}

#if STA_MODE
void setup_select_multicast(int cmd){

    if (cmd) slowdown=0;
    else slowdown=1000;

    copy_multicast_is_broadcasting=cmd;
    if (cmd>2) cmd-=3;
    stopMulticast();
    switch (cmd){
        case 0: break;
        case MULTICAST_TR: beginMulticastTransmitter(); break; // 1
        case MULTICAST_RV: beginMulticastReceiver(); break; // 2

        //// case 3: // disable. resent.
        //// case (MULTICAST_TR+3): // 4, resent.
        //// case (MULTICAST_RV+3): // 5, resent.
    }
    multicast_is_broadcasting=cmd;
}
#endif // STA_MODE

void my_setup_http_server(){

    uri_list+=\
        "/s\r\n"\
        "/version\r\n"\
        "/stop (stop the timer)\r\n"\
        "/poweroff\r\n"\
        "/ls\r\n"\
        "/del?filename=<name>\r\n"\
        "/cp?src=<src_filename>&dst=<dst_filename>\r\n"\
        "/reset\r\n"\
        "/multicast?cmd=<0(disable) 1(as transmitter) 2(as receiver). for sta mode only, multicast/8888/udp>\r\n"\
        "/g1 (right)\r\n"\
        "/g2 (up)\r\n"\
        "/g3 (left)\r\n"\
        "/g4 (down)\r\n"\
        "/g5 (push button)\r\n";

    server_p->on("/s", []() {
        my_calculation_s();
        server_p->send(200, "text/plain", speed_str.c_str());
    });

    server_p->on("/g1", []() {
        joystick_tri[1]+=((joystick_tri[1]&0x1)? 1: 2); // right
        server_p->send(200, "text/plain", "1");
    });

    server_p->on("/g2", []() {
        joystick_tri[0]+=((joystick_tri[0]&0x1)? 1: 2); // up
        server_p->send(200, "text/plain", "2");
    });

    server_p->on("/g3", []() {
        joystick_tri[1]+=((joystick_tri[1]&0x1)? 2: 1); // left
        server_p->send(200, "text/plain", "3");
    });

    server_p->on("/g4", []() {
        joystick_tri[0]+=((joystick_tri[0]&0x1)? 2: 1); // down
        server_p->send(200, "text/plain", "4");
    });

    server_p->on("/g5", []() {
        joystick_tri[2]++; // button, counterclockwise
        server_p->send(200, "text/plain", "5");
    });

    server_p->on("/version", []() { // show version
        server_p->send(200, "text/plain", VERSION_202107190101b);
    });

    server_p->on("/stop", []() { // stop the timer, in order to update local.
        extern unsigned stop_the_timer;
        stop_the_timer=1;
        server_p->send(200, "text/plain", "Done");
    });

    server_p->on("/poweroff", []() {
        extern unsigned poweroff;
        poweroff=1;
        server_p->send(200, "text/plain", "Done");
        
    });

    server_p->on("/ls", []() { // list directory
        String str="";
        cDataBase::Ls("/", &str);
        server_p->send(200, "text/plain", str.c_str());
    });

    server_p->on("/del", []() { // to delete the specified file.
        String fn=server_p->arg("filename");
        if (fn=="") server_p->send(404, "text/plain", "0. page not found.");
        else {
            if (cDataBase((String("/")+fn).c_str()).Remove())
                server_p->send(200, "text/plain", (String("1. removed: ")+fn).c_str());
            else server_p->send(404, "text/plain", "0. file not found.");
        }
    });

    server_p->on("/cp", []() { // to copy files.
        String src=server_p->arg("src");
        String dst=server_p->arg("dst");
        if (src=="") server_p->send(404, "text/plain", "0. page not found.");
        else if (dst=="") server_p->send(404, "text/plain", "0. page not found.");
        else {
            src="/"+src;
            dst="/"+dst;
            cDataBase::Cp(src.c_str(), dst.c_str());
            server_p->send(200, "text/plain", "1. copied.");
        }
    });

    server_p->on("/reset", []() {
        server_p->send(200, "text/plain", "now reset...");
        timer_draw_manip.Pause();
        DrawEnd(draw_page_1);
        DrawConfig(draw_page_1, loop_u8g2_bgblack);
        demandcall_draw_manip();
        printf("\r\n[reset]\r\n");
        delay(1000);
        pinMode(MUX_S1_GPIO, INPUT);
        digitalWrite(BUTTON_GPIO, LOW);
        pinMode(BUTTON_GPIO, INPUT);
        system_deep_sleep(0);
        while (1);
    });

#if STA_MODE
    server_p->on("/multicast", []() {
        // http://<ip address>/multicast?cmd=<0, disable or 1, tr or 2, rv or 3, disable-resent or 4, tr-resent or 5, rv-resent>.
        int cmd=String(server_p->arg("cmd")).toInt();
        if (cmd<0 || cmd>5){
            server_p->send(404, "text/plain", "0. page not found.");
            return;
        }
        setup_select_multicast(cmd);
        server_p->send(200, "text/plain", (String("1. accepted: ")+cmd).c_str());
    });
#endif // STA_MODE

}

void InputLoadHistory(){
    cDataBase s("/history"); // the format likes "abc\ndef\n...xyz\n".
    unsigned char u[8];
    input_str_history="";
    s.Open();

    unsigned x=s.ReadData(u, 7, "history", 0);
    unsigned y=x;
    if (y>7) y=7;
    while (1){
        u[y]=0;
        input_str_history+=(char*)u;
        x-=y;
        if (x<=7){
            if (!x) break;
            y=s.ReadDataCont(u, x);
        }
        else y=s.ReadDataCont(u, 7);
    }

    s.Close();

    input_str_history="Cancel. The current input<"+input_str+">\n"+input_str_history_notsave+input_str_history+"<--EOF\n";
}

bool InputSaveHistory(){ // input string can not have the "\n".

#if 0 // not used
    // from the impl., input string can not have the "\n".
    input_str_history=\
        input_str_history.substring(\
            input_str_history.indexOf('\n')+1,\
            input_str_history.lastIndexOf('\n', input_str_history.length()-2)\
        );
    input_str_history="\n"+input_str()+"\n";
#endif // not used

    if (!input_str.length()) return false;
    cDataBase s("/history"); // the format likes "abc\ndef\n...xyz\n".
    unsigned char u[8];
    input_str_history="";
    s.Open();

    unsigned x=s.ReadData(u, 7, "history", 0);
    unsigned y=x;
    if (y>7) y=7;
    while (1){
        u[y]=0;
        input_str_history+=(char*)u;
        x-=y;
        if (x<=7){
            if (!x) break;
            y=s.ReadDataCont(u, x);
        }
        else y=s.ReadDataCont(u, 7);
    }

    s.Close();

    timer_draw_manip.Pause(); delay(250);

    input_str_history=input_str+"\n"+input_str_history;
    s.ReplaceBytesEx("history", 0, input_str_history.c_str(), input_str_history.length(), input_str_history.length());

    timer_draw_manip.Resume(); delay(250);
    return true;
}

void init_params(){
    cDataBase s("/settings"), h("/history");
    h.Open(); h.Close(); // in order to create a new file.
    s.Open();
    unsigned BootTimes=0;

    if (s.ReadWords(BootTimes, "BootTimes", 0)){
        printf("load params success: %s %u\r\n", "BootTimes", BootTimes);
    }

    if (!BootTimes){
        String str;
        for (int m=0, n=sizeof(input_str_history_default)/sizeof(*input_str_history_default); m<n; m++)
            str=(str+input_str_history_default[m]+"\n");
        h.ReplaceBytesEx("history", 0, str.c_str(), str.length(), str.length());
    }

    s.ReadWords(copy_sta_do_the_connect, "copy_sta_do_the_connect", 0);
    s.ReadWords(pwron_spd_conn, "pwron_spd_conn", 0);
    s.ReadWords(pwron_keep_alive, "pwron_keep_alive", 0);

    s.Close();
    s.ReplaceWordsEx("BootTimes", 0, ++BootTimes);
    if (!((BootTimes+1)%5)){
        s.Renew();
        h.Renew();
    }

    // [[main controls]]
    if (pwron_spd_conn) copy_sta_do_the_connect=1;
}


void setup(){
    // the order in setup() is of concern.
    delay(2000); // the delay 2 seconds is must by sil usb2ur chip.
    Serial.begin(115200);
    while (!Serial) delay(5);
    printf("\r\nwelcome\r\n\r\n\r\n");
    Serial.flush();

    setup_u8g2();
    DrawSetup();
    DrawConfig(draw_page_1, loop_u8g2_welcome);
    DrawSubmit=1;
    demandcall_draw_manip();

    ////pinMode(BUTTON_GPIO, INPUT_PULLUP); // not effective.
    pinMode(BUTTON_GPIO, OUTPUT);
    digitalWrite(BUTTON_GPIO, HIGH); // keep the power
    pinMode(MUX_S0_GPIO, OUTPUT); // mux gpio s0
    pinMode(MUX_S1_GPIO, OUTPUT); // mux gpio s1

    randomSeed(micros()%ESP.getChipId());
    sta_do_the_connect=0; // do not connect ap as default.
    copy_sta_do_the_connect=0; // do not connect ap as default. it is changed in init_params().
    pwron_keep_alive=0; // default 0, no auto keep alive. it is changed in init_params().
    init_params();
    setup_IoT();

    // ------ config the main menu
    if (!copy_sta_do_the_connect) delay(250);
    DrawEnd(draw_page_1);
    DrawConfig(draw_page_1, loop_u8g2_u8g2ani);
    DrawConfig(draw_page_1, loop_u8g2_joystick);
    DrawConfig(draw_page_1, loop_u8g2_status_051);
    DrawConfig(draw_page_1, loop_u8g2_status_052);
    DrawConfig(draw_page_1, loop_u8g2_status_053);
    DrawConfig(draw_page_1, loop_u8g2_status_054);
    DrawConfig(draw_page_1, loop_u8g2_status_055);
    AutoDrawSubmit=1;
    timer_draw_manip.setTimer(BTN_PERIOD, isr_draw_manip_timer);
    timer_draw_manip.Resume();
}


IRAM_ATTR const char *bat_status="-"; // (-)=0, (f)ull=1, (d)ischarging=2, (c)harging=3, (l)ow=4.
void checkBattery(){ // better polling at 10 seconds
    static unsigned f_tmo=0;
    static unsigned battery_adc=joystick[3]; // the current battery ADC value
    if (*bat_status=='l') return; // 4
    if (battery_adc<joystick[3]){bat_status="c"; f_tmo=0;} // 3
    else if (battery_adc>joystick[3]){bat_status="d"; f_tmo=0;} // 2
    else {
        if (!f_tmo) f_tmo=micros();
        else if (Timeout(50000, f_tmo)){ // 50 seconds no change means full
            bat_status="f"; // 1
        }
    }
    battery_adc=joystick[3];
}


void loop(){

    // ------ iot processing. to reconnect is monopolized. so disable the timer first
    loop_IoT();
#if STA_MODE
    static unsigned iflag=0;
    if (!iflag && (is_station_mode_established!=copy_sta_do_the_connect)){
        timer_draw_manip.Pause();
        iflag=1;
        delay(300);
        sta_do_the_connect=copy_sta_do_the_connect;
    }
    else if (iflag && (is_station_mode_established==sta_do_the_connect)){
        timer_draw_manip.Resume();
        iflag=0;
        delay(300);
        if (is_station_mode_established){
            netstr_ssid=WiFi.SSID();
            netstr_remoteip=remote_ap_ip;
            netstr_localip=WiFi.localIP().toString();
            strcpy(http_request_ip, remote_ap_ip); //// assume the ap is the server to http request.
            web_link=String("http://")+http_request_ip+"/s"; //// later on, could be changed by "motor control->find target".
        }
    }
    static unsigned abort_conn=0;
    if (is_station_mode_established!=sta_do_the_connect){ // need to abort if always failed; after 30 seconds.
        static unsigned t=0;
        if (!abort_conn){
            abort_conn=30000; // 30 seconds
            t=micros();
        }
        else if (Timeout(abort_conn, t, true, true)){
            sta_do_the_connect^=1;
            copy_sta_do_the_connect=sta_do_the_connect;
            abort_conn=0;
        }
        return;
    }
    else abort_conn=0;
#endif // STA_MODE

    // ------ button sensing then draw processing
    static unsigned tmo4=micros();
    if (Timeout(75, tmo4)){ // adjust the menu and key response time from 200 to below 75.
        DrawProcessing();
    }

    // ------ check the battery status at each 20 seconds
    static unsigned tmo3=micros();
    if (Timeout(20000, tmo3)){
        checkBattery();
        if (*bat_status=='l'){
            float a=BVOLTAGE(joystick[3]);
            if ((a<3.45f) && !countdown) countdown=9;
            else if (a<3.65f) DrawHoldOnEx_MsgBoxShowing("  BATTERY LOW!");
        }
    }

    // ------ stop the timer in order to webupdate
    if (stop_the_timer){
        static unsigned timer_tmo;
        if (stop_the_timer==1){
            stop_the_timer=2;
            timer_draw_manip.Stop();
            pinMode(MUX_S1_GPIO, INPUT);
        }
        else if (stop_the_timer==3){
            stop_the_timer=4;
            timer_draw_manip.Pause();
            timer_tmo=micros();
        }
        else if ((stop_the_timer==4) && Timeout(60000, timer_tmo)){
            stop_the_timer=0;
            timer_draw_manip.Resume();
        }
    }

#if STA_MODE
    // ------ [[main controls]] power on speed control
    static unsigned t_pwron_spd_conn=0;
    if (pwron_spd_conn){
        joystick_alive=1; // keep the alive
        if (joystick_tri[0] || joystick_tri[1] || joystick_tri[2]) pwron_spd_conn=0;
        else if (!mapx_idx){
            mapx_idx=2; // page_u8g2_speed
            DrawEnd(draw_page_1);
            DrawConfig(draw_page_1, loop_u8g2_speed);
        }
        else if (t_pwron_spd_conn){ // means time countdown
            if (Timeout(15000, t_pwron_spd_conn)){ // reconnect after 15 seconds after auto abort elsewhere
                copy_sta_do_the_connect=1;
                t_pwron_spd_conn=0;
            }
        }
        else if (!copy_sta_do_the_connect){ // means auto abort elsewhere
            t_pwron_spd_conn=micros();
        }
    }
#endif // STA_MODE

#if STA_MODE
    // ------ [[main controls]] once demand, periodically queries whether specific services exist.
    // in this instance, "multicast, 8888" is taken and when the local device assigned as the off=3/tr=4/rv=5(assigned by http or by code),
    // it will start to search the wanted service devices and ask them to be disable/receiver/transmitters.
    // stop search at a round if found at least one and http request only once. reassign can retrigger if needed.
    if ((copy_multicast_is_broadcasting>2) && is_station_mode_established){
        static unsigned tmo5=0;
        if (!tmo5) tmo5=micros()-30000000;
        else if (Timeout(30000, tmo5)){ // set the period as 30 seconds.
            if (int n=mDnsQueryStart("multicast", "udp")){
                int m=0;
                int p=(multicast_is_broadcasting==MULTICAST_TR)? MULTICAST_RV: MULTICAST_TR;
                if (copy_multicast_is_broadcasting==3) p=0;
                for (int i=0; i<n; i++){
                    IPAddress ip;
                    unsigned port;
                    String hname;
                    mDnsQueryCont(i, ip, port, hname);
                    printf("[%d found][%s][%s:%u]\r\n", n, hname.c_str(), ip.toString().c_str(), port);
                    delay(1500); // need to delay for http request more probably success.
                    if (HTTPC_http_client_connect(("http://"+hname+"/multicast?cmd="+p).c_str())){
                        multicast_found+=hname+",";
                        m++;
                    }
                }
                multicast_tr_count=m;
                mDnsQueryEnd();
                tmo5=0;
                copy_multicast_is_broadcasting=-1;
            }
            else { // since services enumerated takes time and might failed/lost datagram.
                multicast_found="";
                multicast_tr_count="0";
                printf("Services not found yet.\r\n");
            }
        }
    }
#endif // STA_MODE

#if STA_MODE
    // ------ [[main controls]] once connection established, periodically queries whether "mngcenter" service exists.
    // in this instance, "mngcenter, 80" is taken which is the real target this display device wants to retrieve real speed data.
    // and the web_link is also modified here.
    // stop search if found any.
    static unsigned tmo6=0, mngcenter_is_exist=0;
    if (is_station_mode_established){
        if (!mngcenter_is_exist){
            if (!tmo6) tmo6=micros()-20000000;
            else if (Timeout(30000, tmo6)){ // set the period as 30 seconds.
                if (mDnsQueryStart("mngcenter", "tcp")){
                    IPAddress ip;
                    unsigned port;
                    String hname;
                    mDnsQueryCont(0, ip, port, hname);
                    printf("[mngcenter found][%s][%s:%u]\r\n", hname.c_str(), ip.toString().c_str(), port);
                    mDnsQueryEnd();

                    netstr_mngcenter_ip=ip.toString();
                    netstr_mngcenter_web="http://"+hname+".local/ip";
                    strcpy(http_request_ip, netstr_mngcenter_ip.c_str());
                    web_link="http://"+netstr_mngcenter_ip+"/s";
                    setup_select_multicast(MULTICAST_RV); // the role it ought to be.
                    only_pause_the_auto_http_req=false;

                    tmo6=0;
                    mngcenter_is_exist=1;
                }
                else { // since services enumerated takes time and might failed/lost datagram.
                    printf("mngcenter not found yet.\r\n");
                }
            }
        }
    }
    else if (mngcenter_is_exist){mngcenter_is_exist=0; netstr_mngcenter_ip=""; setup_select_multicast(0);}
#endif // STA_MODE

    // ------ after 60 seconds to power down if idle. note that whenever webupdate, press a button to get free time or keep it alive.
    static unsigned tmo2=micros(), tmo22=tmo2;
    if (pwron_keep_alive && !countdown);
    else if (joystick_alive){
        joystick_alive=0;
        countdown=0;
        tmo22=tmo2=micros();
    }
    else if (Timeout(50000, tmo2, true) && !countdown) countdown=9;
    else if (countdown){
        if (Timeout(1050, tmo22)) countdown--;
    }
    if ((countdown==1) || poweroff){
        timer_draw_manip.Pause();
        DrawEnd(draw_page_1);
        DrawConfig(draw_page_1, loop_u8g2_bgblack);
        demandcall_draw_manip();
        printf("\r\n[power down]\r\n");
        delay(1000);
        pinMode(MUX_S1_GPIO, INPUT);
        digitalWrite(BUTTON_GPIO, LOW);
        pinMode(BUTTON_GPIO, INPUT);
        system_deep_sleep(0);
        while (1);
    }

    // ------ 7 seconds timeout if motor command did not proceeding.
    static unsigned tmo0=0;
    if (motor_issued_cmd_state && motor_issued_cmd_state<=3){
        if (!tmo0) tmo0=micros();
        else if (Timeout(7000, tmo0)){
            if (DrawHoldOnEx_isdoneMsgBoxShowing()){
                DrawHoldOnEx_MsgBoxShowing("check connection");
                motor_issued_cmd_state=0;
                tmo0=0;
            }
        }
    }
    else tmo0=0;

    // ------ debug
    /*static unsigned tmo1=micros();
    if (Timeout(1500, tmo1)){
        printf("%d %d %d %d %d[x[%d, %d], y[%d, %d], b[%d, %d], %d[%d, %fV]]\r\n",\
        joystick_dir, slide_2_pos, joystick_action, jtk_pos_x, jtk_pos_y,\
        joystick[1], joystick_tri[1],\
        joystick[0], joystick_tri[0],\
        joystick[2], joystick_tri[2],\
        joystick[3], joystick_tri[3], BVOLTAGE(joystick[3]));
    }*/
}
